import os

import numpy as np
import pandas as pd
import scipy.sparse as ssp
import scipy.stats as stats
import seaborn as sns
import statsmodels.sandbox.stats.multicomp
from ete3 import Tree
from matplotlib import pyplot as plt
from scipy.cluster import hierarchy
from sklearn.manifold import SpectralEmbedding

from .. import help_functions as hf
from .. import logging as logg
from .. import settings
from .. import tool as tl


def darken_cmap(cmap, scale_factor):
    """
    Generate a gradient color map for plotting.
    """

    cdat = np.zeros((cmap.N, 4))
    for ii in range(cdat.shape[0]):
        curcol = cmap(ii)
        cdat[ii, 0] = curcol[0] * scale_factor
        cdat[ii, 1] = curcol[1] * scale_factor
        cdat[ii, 2] = curcol[2] * scale_factor
        cdat[ii, 3] = 1
    cmap = cmap.from_list(cmap.N, cdat)
    return cmap


def start_subplot_figure(n_subplots, n_columns=5, fig_width=14, row_height=3):
    """
    Generate a figure object with given subplot specification.
    """

    n_rows = int(np.ceil(n_subplots / float(n_columns)))
    fig = plt.figure(figsize=(fig_width, n_rows * row_height))
    return fig, n_rows, n_columns


def embedding_genes(adata, basis="X_emb", color=None, color_bar=True, **kwargs):
    """
    A test embedding method for plotting genes

    Works better with subplots
    """
    if basis not in adata.obsm.keys():
        raise ValueError(f"basis={basis} is not among {adata.obsm.keys()}")
    if color in adata.var_names:
        vector = adata[:, color].X.A.flatten()

    elif color in adata.obs.keys():
        vector = np.array(adata.obs[color])
    else:
        raise ValueError(f"color value is not right")

    x_emb = adata.obsm[basis][:, 0]
    y_emb = adata.obsm[basis][:, 1]
    customized_embedding(
        x_emb,
        y_emb,
        vector,
        title=color,
        col_range=(0, 99.8),
        color_bar=color_bar,
        **kwargs,
    )


def embedding(
    adata, basis="X_emb", color=None, cmap=darken_cmap(plt.cm.Reds, 0.9), **kwargs
):
    """
    Scatter plot for user-specified embedding basis.

    We imported :func:`~scanpy.pl.embedding` for this purpose.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Assume to contain transition maps at adata.uns.
    basis: `str`, optional (default: 'X_emb')
        The embedding to use for the plot.
    color: `str, list of str, or None` (default: None)
        Keys for annotations of observations/cells or variables/genes,
        e.g., 'state_info', 'time_info',['Gata1','Gata2']
    """

    from scanpy.plotting._tools import embedding as sc_embedding

    sc_embedding(adata, basis=basis, color=color, cmap=cmap, **kwargs)


def customized_embedding(
    x,
    y,
    vector,
    title=None,
    ax=None,
    order_points=True,
    set_ticks=False,
    col_range=None,
    buffer_pct=0.03,
    point_size=1,
    color_map=None,
    smooth_operator=None,
    set_lim=True,
    vmax=None,
    vmin=None,
    color_bar=False,
    color_bar_label="",
    color_bar_title="",
):
    """
    Plot a vector on an embedding.

    Parameters
    ----------
    x: `np.array`
        x coordinate of the embedding
    y: `np.array`
        y coordinate of the embedding
    vector: `np.array`
        A vector to be plotted.
    color_map: {plt.cm.Reds,plt.cm.Blues,...}, (default: None)
    ax: `axis`, optional (default: None)
        An external ax object can be passed here.
    order_points: `bool`, optional (default: True)
        Order points to plot by the gene expression
    col_range: `tuple`, optional (default: None)
        The default setting is to plot the actual value of the vector.
        If col_range is set within [0,100], it will plot the percentile of the values,
        and the color_bar will show range [0,1]. This re-scaling is useful for
        visualizing gene expression.
    buffer_pct: `float`, optional (default: 0.03)
        Extra space for the plot box frame
    point_size: `int`, optional (default: 1)
        Size of the data point
    smooth_operator: `np.array`, optional (default: None)
        A smooth matrix to be applied to the subsect of gene expression matrix.
    set_lim: `bool`, optional (default: True)
        Set the plot range (x_limit, and y_limit) automatically.
    vmax: `float`, optional (default: np.nan)
        Maximum color range (saturation).
        All values above this will be set as vmax.
    vmin: `float`, optional (default: np.nan)
        The minimum color range, all values below this will be set to be vmin.
    color_bar: `bool`, optional (default, False)
        If True, plot the color bar.
    set_ticks: `bool`, optional (default, False)
        If False, remove figure ticks.

    Returns
    -------
    ax:
        The figure axis
    """

    from matplotlib.colors import Normalize as mpl_Normalize

    if color_map is None:
        color_map = darken_cmap(plt.cm.Reds, 0.9)
    if ax is None:
        fig, ax = plt.subplots()

    coldat = vector.astype(float)

    if smooth_operator is None:
        coldat = coldat.squeeze()
    else:
        coldat = np.dot(smooth_operator, coldat).squeeze()

    if order_points:
        o = np.argsort(coldat)
    else:
        o = np.arange(len(coldat))

    if vmin is None:
        if col_range is None:
            vmin = np.min(coldat)
        else:
            vmin = np.percentile(coldat, col_range[0])

    if vmax is None:
        if col_range is None:
            vmax = np.max(coldat)
        else:
            vmax = np.percentile(coldat, col_range[1])

    if vmax == vmin:
        vmax = coldat.max()

    ax.scatter(
        x[o], y[o], c=coldat[o], s=point_size, cmap=color_map, vmin=vmin, vmax=vmax
    )

    if not set_ticks:
        ax.set_xticks([])
        ax.set_yticks([])
        ax.axis("off")

    if set_lim == True:
        ax.set_xlim(x.min() - x.ptp() * buffer_pct, x.max() + x.ptp() * buffer_pct)
        ax.set_ylim(y.min() - y.ptp() * buffer_pct, y.max() + y.ptp() * buffer_pct)

    if title is not None:
        ax.set_title(title)

    if color_bar:

        norm = mpl_Normalize(vmin=vmin, vmax=vmax)
        Clb = plt.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=color_map), ax=ax)
        Clb.set_label(
            color_bar_label,
            rotation=270,
            labelpad=20,
        )
        Clb.ax.set_title(color_bar_title)
    return ax


def plot_neighbor_joining(
    output_directory,
    node_groups,
    X_history,
    merged_pairs_history,
    node_names_history,
):
    fig, axs = plt.subplots(1, len(X_history))
    for i, X in enumerate(X_history):
        vmax = 1.2 * np.max(np.triu(X, k=1))
        axs[i].imshow(X, vmax=vmax)
        ii, jj = merged_pairs_history[i]
        axs[i].scatter([jj], [ii], s=100, marker="*", c="white")

        column_groups = [node_groups[n] for n in node_names_history[i]]
        column_labels = [" + ".join([n for n in grp]) for grp in column_groups]
        axs[i].set_xticks(np.arange(X.shape[1]) + 0.2)
        axs[i].set_xticklabels(column_labels, rotation=90, ha="right")
        axs[i].set_xlim([-0.5, X.shape[1] - 0.5])
        axs[i].set_ylim([X.shape[1] - 0.5, -0.5])
        axs[i].set_yticks(np.arange(X.shape[1]) + 0.2)
        axs[i].set_yticklabels(["" for grp in column_groups], rotation=90, ha="right")
    fig.set_size_inches((16, 4))
    plt.savefig(os.path.join(output_directory, "neighbor_joint_heatmaps.pdf"))


def custom_hierachical_ordering(order_ids, matrix, pseudo_count=0.00001):
    """
    A recursive algorithm to rank the clones.

    The matrix is fate-by-clone, and we order it in the clone dimension
    """
    if (len(order_ids) < 1) or (matrix.shape[1] < 2):
        return matrix

    order_ids = np.array(order_ids)
    valid_clone_idx = np.ones(matrix.shape[1]) > 0
    new_data_list = []
    for j, x in enumerate(order_ids):
        valid_clone_idx_tmp = valid_clone_idx & (matrix[x] > 0)
        data_matrix = matrix[:, valid_clone_idx_tmp].T
        valid_clone_idx = valid_clone_idx & (~valid_clone_idx_tmp)
        if np.sum(valid_clone_idx_tmp) >= 2:
            order_y = hf.get_hierch_order(data_matrix + pseudo_count)
            updated_matrix = data_matrix[order_y].T
        else:
            updated_matrix = data_matrix.T
        updated_matrix_1 = custom_hierachical_ordering(
            order_ids[j + 1 :], updated_matrix, pseudo_count=pseudo_count
        )
        new_data_list.append(updated_matrix_1)
    new_data_list.append(
        matrix[:, valid_clone_idx]
    )  # add the remaining clones not selected before
    return np.column_stack(new_data_list)


def heatmap(
    data_matrix,
    order_map_x=True,
    order_map_y=True,
    x_ticks=None,
    y_ticks=None,
    x_tick_style=None,
    y_tick_style=None,
    col_range=[0, 99],
    color_bar_label="",
    log_transform=False,
    color_map=plt.cm.Reds,
    vmin=None,
    vmax=None,
    fig_width=4,
    fig_height=6,
    color_bar=True,
    x_label=None,
    y_label=None,
    x_label_style=None,
    y_label_style=None,
    print_ordered_labels=False,
    pseudo_count=10 ** (-10),
    ax=None,
):
    """
    Plot ordered heat map of non-square data_matrix matrix

    Parameters
    ----------
    data_matrix: `np.array`
        The data matrix to be plotted
    order_map_x: `bool`
        Whether to re-order the x coordinate of the matrix or not
    order_map_y: `bool`
        Whether to re-order the y coordinate of the matrix or not
    x_ticks, y_ticks: `list`
        List of variable names for x and y ticks
    color_bar_label: `str`, optional (default: 'cov')
        Color bar label
    data_des: `str`, optional (default: '')
        String to distinguish different saved objects.
    log_transform: `bool`, optional (default: False)
        If true, perform a log transform. This is needed when the data
        matrix has entries varying by several order of magnitude.
    col_range: `tuple`, optional (default: None)
        The default setting is to plot the actual value of the vector.
        If col_range is set within [0,100], it will plot the percentile of the values,
        and the color_bar will show range [0,1]. This re-scaling is useful for
        visualizing gene expression.

    Returns
    ax,
    order_x,
        An integer index for ordered x axis
    order_y,
        An integer index for ordered y axis
    """

    from matplotlib.colors import Normalize as mpl_Normalize

    x_array = np.arange(data_matrix.shape[1])
    y_array = np.arange(data_matrix.shape[0])
    if order_map_x and (data_matrix.shape[1] > 2):
        if data_matrix.shape[0] != data_matrix.shape[1]:
            X = tl.get_normalized_covariance(data_matrix + pseudo_count, method="SW")
            Z = hierarchy.ward(X)
            order_x = hierarchy.leaves_list(hierarchy.optimal_leaf_ordering(Z, X))
        else:
            Z = hierarchy.ward(data_matrix + pseudo_count)
            order_x = hierarchy.leaves_list(
                hierarchy.optimal_leaf_ordering(Z, data_matrix + pseudo_count)
            )
    else:
        order_x = x_array

    if order_map_y and (data_matrix.shape[0] > 2):
        if data_matrix.shape[0] != data_matrix.shape[1]:
            order_y = hf.get_hierch_order(data_matrix + pseudo_count)
        else:
            Z = hierarchy.ward(data_matrix + pseudo_count)
            order_y = hierarchy.leaves_list(
                hierarchy.optimal_leaf_ordering(Z, data_matrix + pseudo_count)
            )
    else:
        order_y = y_array

    if log_transform:
        new_data = np.log(data_matrix[order_y][:, order_x] + 1) / np.log(10)
        label_ = " (log10)"
    else:
        new_data = data_matrix[order_y][:, order_x]
        label_ = ""

    col_data = new_data.flatten()
    if vmin is None:
        if col_range is None:
            vmin = np.min(col_data)
        else:
            vmin = np.percentile(col_data, col_range[0])

    if vmax is None:
        if col_range is None:
            vmax = np.max(col_data)
        else:
            vmax = np.percentile(col_data, col_range[1])
            if (vmax == 0) & (np.max(col_data) >= 1):
                vmax = 1
            if (vmax == 0) & (np.max(col_data) <= 1):
                vmax = np.max(col_data)

    if ax is None:
        fig, ax = plt.subplots()
        ax_ = None
    else:
        ax_ = ax
    ax.imshow(
        new_data,
        aspect="auto",
        cmap=color_map,
        vmin=vmin,
        vmax=vmax,
        interpolation="none",
    )

    if x_ticks is None:
        plt.xticks([])
    else:
        plt.xticks(
            x_array,
            np.array(x_ticks)[order_x],
            rotation=90,
            style=x_tick_style,
        )

        if print_ordered_labels:
            x_ticks_print = "','".join(np.array(x_ticks)[order_x])
            print(f"x_order: [{','.join(np.array(order_x).astype(str))}]")
            print(f"x_ticks: ['{x_ticks_print}']")

    if y_ticks is None:
        plt.yticks([])
    else:
        plt.yticks(
            y_array,
            np.array(y_ticks)[order_y],
            style=y_tick_style,
        )

        if print_ordered_labels:
            y_ticks_print = "','".join(np.array(y_ticks)[order_y])
            print(f"y_order: [{','.join(np.array(order_y).astype(str))}]")
            print(f"y_ticks: ['{y_ticks_print}']")

    if x_label is not None:
        ax.set_xlabel(x_label, style=x_label_style)

    if y_label is not None:
        ax.set_ylabel(y_label, style=y_label_style)

    if color_bar:
        norm = mpl_Normalize(vmin=vmin, vmax=vmax)
        cbar = plt.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=color_map))
        cbar.set_label(f"{color_bar_label}{label_}", rotation=270, labelpad=20)
    if ax_ is None:
        plt.gcf().set_size_inches((fig_width, fig_height))
    return ax


def fate_map_embedding(
    adata,
    fate_vector,
    cell_id_t1,
    sp_idx,
    mask=None,
    target_list=None,
    color_bar_label="",
    color_bar_title="",
    figure_title="",
    background=True,
    show_histogram=False,
    auto_color_scale=True,
    color_bar=True,
    horizontal=False,
    target_transparency=0.2,
    histogram_scales=None,
    color_map=None,
    order_method=None,
    vmax=None,
    vmin=None,
):
    """
    Note: sp_idx is a bool array, of the length len(cell_id_t1)
    mask: bool array  of length adata.shape[0]
    """

    fig_width = settings.fig_width
    fig_height = settings.fig_height
    point_size = settings.fig_point_size

    x_emb = adata.obsm["X_emb"][:, 0]
    y_emb = adata.obsm["X_emb"][:, 1]
    state_info = np.array(adata.obs["state_info"])

    if mask is not None:
        if len(mask) == adata.shape[0]:
            mask = mask.astype(bool)
            sp_idx = sp_idx & (mask[cell_id_t1])
        else:
            logg.error("mask length does not match adata.shape[0]. Ignored mask.")

    if np.sum(sp_idx) == 0:
        raise ValueError("No cells selected")

    if color_bar:
        fig_width = fig_width + 1

    if show_histogram:
        tot_N = 2
    else:
        tot_N = 1

    if horizontal:
        row = 1
        col = tot_N
    else:
        row = tot_N
        col = 1

    fig = plt.figure(figsize=(fig_width * col, fig_height * row))

    fate_map_temp = fate_vector[cell_id_t1][sp_idx]
    ax0 = plt.subplot(row, col, 1)
    if background:
        customized_embedding(
            x_emb,
            y_emb,
            np.zeros(len(y_emb)),
            point_size=point_size,
            ax=ax0,
            title=figure_title,
        )
    else:
        customized_embedding(
            x_emb[cell_id_t1][sp_idx],
            y_emb[cell_id_t1][sp_idx],
            np.zeros(len(y_emb[cell_id_t1][sp_idx])),
            point_size=point_size,
            ax=ax0,
            title=figure_title,
        )

    if target_list is not None:
        for zz in target_list:
            idx_2 = state_info == zz
            ax0.plot(
                x_emb[idx_2],
                y_emb[idx_2],
                ".",
                color="cyan",
                markersize=point_size * 1,
                alpha=target_transparency,
            )

    if auto_color_scale:
        vmax = None
        vmin = None
    else:
        if vmax is None:
            vmax = 1
        if vmin is None:
            vmin = 0

    if order_method == "fate_bias":
        new_idx = np.argsort(abs(fate_map_temp - 0.5))
    else:
        # new_idx = np.arange(len(fate_map_temp))
        new_idx = np.argsort(abs(fate_map_temp))

    customized_embedding(
        x_emb[cell_id_t1][sp_idx][new_idx],
        y_emb[cell_id_t1][sp_idx][new_idx],
        fate_map_temp[new_idx],
        point_size=point_size,
        ax=ax0,
        title=figure_title,
        set_lim=False,
        vmax=vmax,
        vmin=vmin,
        color_bar=color_bar,
        color_bar_label=color_bar_label,
        color_bar_title=color_bar_title,
        color_map=color_map,
        order_points=False,
    )

    if show_histogram:
        ax = plt.subplot(row, col, 2)
        ax.hist(fate_map_temp, 50, color="#2ca02c", density=True)
        if histogram_scales is not None:
            ax.set_xlim(histogram_scales)
        ax.set_xlabel(color_bar_label)
        ax.set_ylabel("Density")
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.set_title(f"Ave.: {np.mean(fate_map_temp):.2f}")


def rand_jitter(arr, fraction):
    stdev = fraction * np.mean(arr)
    # stdev = fraction * (max(arr) - min(arr))
    return arr + np.random.randn(len(arr)) * stdev


def jitter(
    x,
    y,
    s=10,
    c="b",
    marker="o",
    cmap=None,
    norm=None,
    vmin=None,
    vmax=None,
    alpha=None,
    linewidths=None,
    std=0.01,
    **kwargs,
):
    return plt.scatter(
        rand_jitter(x, std),
        rand_jitter(y, std),
        s=s,
        c=c,
        marker=marker,
        cmap=cmap,
        norm=norm,
        vmin=vmin,
        vmax=vmax,
        alpha=alpha,
        linewidths=linewidths,
        **kwargs,
    )


def plot_one_cluster(adata, cluster_name=None, cluster_obs_key=None, basis="X_umap"):
    df_plot = pd.DataFrame(
        {
            "cluster": adata.obs[cluster_obs_key],
            "x": adata.obsm[basis][:, 0],
            "y": adata.obsm[basis][:, 1],
        }
    )

    df_plot_tmp_1 = df_plot[df_plot["cluster"] != cluster_name]
    df_plot_tmp_1.loc[:, "cluster"] = "others"
    df_map_v2 = pd.concat([df_plot_tmp_1, df_plot[df_plot["cluster"] == cluster_name]])
    g = sns.relplot(kind="scatter", data=df_map_v2, x="x", y="y", hue="cluster", s=5)
    g.ax.axis("off")


def visualize_tree(
    input_tree,
    color_coding: dict = None,
    mode="r",
    width=60,
    height=60,
    dpi=300,
    data_des="tree",
    figure_path=".",
    line_width=0,
    marker_size_internal=5,
    marker_size_leaf=5,
):
    """
    Visualize a tree structured in ete3 style.

    We provide the option to color code the leaf of the tree. For example, if the leaf nodes represent single cells,
    and some are more similar than others according to other information, we may impose the same color for these similar cells.
    Then, visualization with this color setting will allow us see how the tree structure recaptulate the similarity of these cells. If so, similar colors tend to cluster together.

    This function will require a full installation of the ete3 packages (which are not part of the default cospar installation), including ete3, ete_toolchain, PyQt5, QtPy. Please run the following to install.

    ```bash
    conda install -c etetoolkit ete3 ete_toolchain
    pip install PyQt5
    pip install QtPy
    ```

    Parameters
    ----------
    input_tree:
        A tree stored in ete3 style. This can be the output from running `cs.tl.fate_hierarchy(adata, source="X_clone")`, where the resulting tree will be stored at my_tree = adata.uns["fate_hierarchy_X_clone"]["tree"].
    color_coding:
        A dictionary for mapping the leaf names to a specific color. An example color_coding will be {'node_1':"#e5f5f9",'node_2':"#99d8c9",...}. 'node_1' and 'node_2' will be the leaf node names.
    mode:
        The mode of plotting. {'r', 'c'}. 'c' is the circular mode, and 'r' is the rectangular mode.
    width:
        Width of the tree plot.
    height:
        Height of the tree plot.
    dpi:
        Resolution of the tree plot.
    data_des:
        Label for saving the figure, i.e., figure name.
    figure_path:
        Figure directory.
    """

    from ete3 import AttrFace, NodeStyle, Tree, TreeStyle, faces
    from IPython.display import Image, display

    def layout(node):
        if node.is_leaf():  # this is the part showing the leaf
            N = AttrFace("name", fsize=5)
            faces.add_face_to_node(N, node, 100, position="aligned")
            # pass

    if color_coding is not None:
        print("coding")
        for n in input_tree.traverse():  # internal node
            nst1 = NodeStyle(
                size=marker_size_internal,
                fgcolor="#f0f0f0",
                vt_line_width=line_width,
                hz_line_width=line_width,
            )
            n.set_style(nst1)

        for n in input_tree:
            for key, value in color_coding.items():
                if n.name == key:
                    nst1 = NodeStyle(
                        size=marker_size_leaf,
                        hz_line_width=line_width,
                        fgcolor="#000000",
                    )
                    nst1["bgcolor"] = value
                    n.set_style(nst1)

    ts = TreeStyle()
    # ts.layout_fn = layout # layout not used. It will add faces to each node, and each fates is the leaf name
    ts.mode = mode
    ts.show_leaf_name = False
    # ts.extra_branch_line_color = "red"
    # ts.extra_branch_line_type = 0
    input_tree.render(
        os.path.join(figure_path, f"{data_des}.pdf"),
        tree_style=ts,
        w=width,
        h=height,
        units="mm",
    )
    input_tree.render(
        os.path.join(figure_path, f"{data_des}.png"),
        tree_style=ts,
        w=width,
        h=height,
        dpi=dpi,
        units="mm",
    )

    display(Image(filename=os.path.join(figure_path, f"{data_des}.png")))


def plot_adata_with_prefered_order(
    adata,
    obs_key,
    basis="X_umap",
    plot_order=None,
    palette=None,
    background=None,
    background_color="#d9d9d9",
    background_size=50,
    **kwargs,
):
    """
    An example code
    ```python
    plot_adata_with_prefered_order(adata,'Fate_bias',plot_order=['No Meg bias','Early Meg bias','Late Meg bias' ],
                                  palette={'No Meg bias':'#ff7f0e','Early Meg bias':'#1f77b4','Late Meg bias':'grey'},
                                  linewidth = 0,s=40)
    #plt.tight_layout()
    plt.savefig(f'{figure_path}/X_emb_fate_bias_Meg.pdf')
    ```
    """
    if plot_order is None:
        plot_order = list(adata.obs[obs_key].unique())
    if palette is None:
        palette = dict(
            zip(plot_order, np.array(sns.color_palette().as_hex())[: len(plot_order)])
        )

    df_fate_map = pd.DataFrame(
        {
            obs_key: adata.obs[obs_key],
            "x": adata.obsm[basis][:, 0],
            "y": adata.obsm[basis][:, 1],
        }
    )
    df_list = []

    if background is not None:
        if background in plot_order:
            plot_order.remove(background)
        df_map_bk = df_fate_map[df_fate_map[obs_key] == background]
        ax = sns.scatterplot(
            data=df_map_bk,
            x="x",
            y="y",
            hue=obs_key,
            palette={background: background_color},
            s=background_size,
            edgecolor=background_color,
        )

    for z in plot_order:
        df_list.append(df_fate_map[df_fate_map[obs_key] == z])

    df_map_v2 = pd.concat(df_list, ignore_index=True)
    ax = sns.scatterplot(
        data=df_map_v2,
        x="x",
        y="y",
        hue=obs_key,
        palette=palette,
        **kwargs,
    )
    plt.axis("off")
    plt.legend(loc=[1.05, 0], frameon=False)
    return ax



# -*- coding: utf-8 -*-

import os
import time

import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse as ssp

from cospar.tmap import _tmap_core as tmap_core
from cospar.tmap import _utils as tmap_util

from .. import help_functions as hf
from .. import logging as logg
from .. import settings
from .. import tool as tl


# v1 version, allows to set later time point
def infer_Tmap_from_multitime_clones(
    adata_orig,
    clonal_time_points=None,
    later_time_point=None,
    smooth_array=[15, 10, 5],
    CoSpar_KNN=20,
    sparsity_threshold=0.1,
    intraclone_threshold=0.05,
    normalization_mode=1,
    extend_Tmap_space=False,
    save_subset=True,
    use_full_Smatrix=True,
    trunca_threshold=[0.001, 0.01],
    compute_new=False,
    max_iter_N=5,
    epsilon_converge=0.05,
):
    """
    Infer transition map for clonal data with multiple time points.

    It prepares adata object for cells of targeted time points by
    :func:`cospar.tmap._utils.select_time_points`, generates the similarity matrix
    via :func:`cospar.tmap._utils.generate_similarity_matrix`, and iteratively calls
    the core function :func:`.refine_Tmap_through_cospar` to update
    the transition map.

    * If `later_time_point=None`, the inferred map allows transitions
      between neighboring time points. For example, if
      clonal_time_points=['day1','day2','day3'], then it computes transitions
      for pairs (day1, day2) and (day2, day3), but not (day1, day3).

    * If `later_time_point` is specified, the function produces a map
      between earlier time points and this later time point. For example, if
      `later_time_point='day3`, the map allows transitions for pairs (day1, day3)
      and (day2, day3), but not (day1,day2).

    Parameters
    ------------
    adata_orig: :class:`~anndata.AnnData` object
        Should be prepared from our anadata initialization.
    clonal_time_points: `list` of `str`, optional (default: all time points)
        List of time points to be included for analysis.
        We assume that each selected time point has clonal measurements.
    later_time_points: `list`, optional (default: None)
        If specified, the function will produce a map T between these early
        time points among `clonal_time_points` and the `later_time_point`.
        If not specified, it produces a map T between neighboring clonal time points.
    smooth_array: `list`, optional (default: [15,10,5])
        List of smooth rounds at initial runs of iteration.
        Suppose that it has a length N. For iteration n<N, the n-th entry of
        smooth_array determines the kernel exponent to build the S matrix at the n-th
        iteration. When n>N, we use the last entry of smooth_array to compute
        the S matrix. We recommend starting with more smoothing depth and gradually
        reduce the depth, as inspired by simulated annealing. Data with higher
        clonal dispersion should start with higher smoothing depth. The final depth should
        depend on the manifold itself. For fewer cells, it results in a small KNN graph,
        and a small final depth should be used. We recommend to use a number at
        the multiple of 5 for computational efficiency i.e.,
        smooth_array=[20, 15, 10, 5], or [20,15,10]
    max_iter_N: `int`, optional (default: 5)
        The maximum iterations used to compute the transition map, regardless of epsilon_converge.
    epsilon_converge: `float`, optional (default: 0.05)
        The convergence threshold for the change of map correlations between consecutive iterations.
        This convergence test is activated only when CoSpar has iterated for 3 times.
    CoSpar_KNN: `int`, optional (default: 20)
        The number of neighbors for KNN graph used for computing the
        similarity matrix.
    trunca_threshold: `list`, optional (default: [0.001,0.01])
        Threshold to reset entries of a matrix to zero. The first entry is for
        Similarity matrix; the second entry is for the Tmap.
        This is only for computational and storage efficiency.
    sparsity_threshold: `float`, optional (default: 0.1)
        The relative threshold to remove noises in the updated transition map,
        in the range [0,1].
    intraclone_threshold: `float`, optional (default: 0.05)
        The threshold to remove noises in the demultiplexed (un-smoothed) map,
        in the range [0,1]
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.
    extend_Tmap_space: `bool` optional (default: `False`)
        If true, the initial states for Tmap will include all states at initial time points,
        and the later states for Tmap will include all states at later time points.
        Otherwise, the initial and later state space of the Tmap will be
        restricted to cells with multi-time clonal information
        alone. The latter case speeds up the computation, which is recommended.
        This option is ignored when `later_time_points` is not None.
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_full_Smatrix: `bool`, optional (default: True)
        If true, extract the relevant Smatrix from the full Smatrix defined by all cells.
        This tends to be more accurate. The package is optimized around this choice.
    Compute_new: `bool`, optional (default: False)
        If True, compute Smatrix from scratch, whether it was
        computed and saved before or not. This is activated only when
        `use_full_Smatrix=False`.

    Returns
    -------
    adata: :class:`~anndata.AnnData` object
        Store results at adata.uns['transition_map']
        and adata.uns['intraclone_transition_map']. This adata is different
        from the input adata_orig due to subsampling cells.
    """

    t0 = time.time()
    hf.check_available_clonal_info(adata_orig)
    clonal_time_points_0 = np.array(adata_orig.uns["clonal_time_points"])
    if len(clonal_time_points_0) < 2:
        raise ValueError("There are no multi-time clones. Abort the inference.")

    if clonal_time_points is None:
        clonal_time_points = clonal_time_points_0

    if type(later_time_point) == list:
        later_time_point = later_time_point[0]

    if later_time_point is not None:
        clonal_time_points = list(clonal_time_points) + [later_time_point]
        clonal_time_points = list(set(clonal_time_points))

    hf.check_input_parameters(
        adata_orig,
        later_time_point=later_time_point,
        clonal_time_points=clonal_time_points,
        smooth_array=smooth_array,
        save_subset=save_subset,
    )
    # order the clonal time points
    time_ordering = adata_orig.uns["time_ordering"]
    sel_idx_temp = np.in1d(time_ordering, clonal_time_points)
    clonal_time_points = time_ordering[sel_idx_temp]

    logg.info("------Compute the full Similarity matrix if necessary------")
    data_path = settings.data_path
    if (
        use_full_Smatrix
    ):  # prepare the similarity matrix with all state info, all subsequent similarity will be down-sampled from this one.

        temp_str = "0" + str(trunca_threshold[0])[2:]
        round_of_smooth = np.max(smooth_array)
        data_des = adata_orig.uns["data_des"][0]
        similarity_file_name = os.path.join(
            data_path,
            f"{data_des}_Similarity_matrix_with_all_cell_states_kNN{CoSpar_KNN}_Truncate{temp_str}",
        )
        if not (
            os.path.exists(similarity_file_name + f"_SM{round_of_smooth}.npz")
            and (not compute_new)
        ):
            similarity_matrix_full = tmap_util.generate_similarity_matrix(
                adata_orig,
                similarity_file_name,
                round_of_smooth=round_of_smooth,
                neighbor_N=CoSpar_KNN,
                truncation_threshold=trunca_threshold[0],
                save_subset=save_subset,
                compute_new_Smatrix=compute_new,
            )

    # compute transition map between neighboring time points
    if later_time_point is None:
        logg.info("----Infer transition map between neighboring time points-----")
        logg.info("Step 1: Select time points")
        adata = tmap_util.select_time_points(
            adata_orig,
            time_point=clonal_time_points,
            extend_Tmap_space=extend_Tmap_space,
        )

        logg.info("Step 2: Optimize the transition map recursively")
        tmap_core.infer_Tmap_from_multitime_clones_private(
            adata,
            smooth_array=smooth_array,
            neighbor_N=CoSpar_KNN,
            sparsity_threshold=sparsity_threshold,
            intraclone_threshold=intraclone_threshold,
            normalization_mode=normalization_mode,
            save_subset=save_subset,
            use_full_Smatrix=use_full_Smatrix,
            trunca_threshold=trunca_threshold,
            compute_new_Smatrix=compute_new,
            max_iter_N=max_iter_N,
            epsilon_converge=epsilon_converge,
        )

        if "Smatrix" in adata.uns.keys():
            adata.uns.pop("Smatrix")

        logg.info(f"-----------Total used time: {time.time()-t0} s ------------")
        return adata

    else:
        # compute transition map between initial time points and the later time point
        sel_id = np.nonzero(np.in1d(clonal_time_points, later_time_point))[0][0]
        initial_time_points = clonal_time_points[:sel_id]

        time_info_orig = np.array(adata_orig.obs["time_info"])
        sp_idx = np.zeros(adata_orig.shape[0], dtype=bool)
        all_time_points = list(initial_time_points) + [later_time_point]
        label = "t"
        for xx in all_time_points:
            id_array = np.nonzero(time_info_orig == xx)[0]
            sp_idx[id_array] = True
            label = label + "*" + str(xx)

        adata = adata_orig[sp_idx]
        data_des_orig = adata_orig.uns["data_des"][0]
        data_des_0 = adata_orig.uns["data_des"][-1]
        data_des = (
            data_des_0
            + f"_MultiTimeClone_Later_FullSpace{int(extend_Tmap_space)}_{label}"
        )
        adata.uns["data_des"] = [data_des_orig, data_des]

        time_info = np.array(adata.obs["time_info"])
        time_index_t2 = time_info == later_time_point
        time_index_t1 = ~time_index_t2

        #### used for similarity matrix generation
        Tmap_cell_id_t1 = np.nonzero(time_index_t1)[0]
        Tmap_cell_id_t2 = np.nonzero(time_index_t2)[0]
        adata.uns["Tmap_cell_id_t1"] = Tmap_cell_id_t1
        adata.uns["Tmap_cell_id_t2"] = Tmap_cell_id_t2
        adata.uns["clonal_cell_id_t1"] = Tmap_cell_id_t1
        adata.uns["clonal_cell_id_t2"] = Tmap_cell_id_t2
        adata.uns["sp_idx"] = sp_idx
        data_path = settings.data_path

        transition_map = np.zeros((len(Tmap_cell_id_t1), len(Tmap_cell_id_t2)))
        intraclone_transition_map = np.zeros(
            (len(Tmap_cell_id_t1), len(Tmap_cell_id_t2))
        )

        logg.info(
            "------Infer transition map between initial time points and the later time one------"
        )
        for yy in initial_time_points:

            logg.info(f"--------Current initial time point: {yy}--------")

            logg.info("Step 1: Select time points")
            adata_temp = tmap_util.select_time_points(
                adata_orig, time_point=[yy, later_time_point], extend_Tmap_space=True
            )  # for this to work, we need to set extend_Tmap_space=True, otherwise for different initial time points, the later Tmap_cell_id_t2 may be different

            logg.info("Step 2: Optimize the transition map recursively")
            tmap_core.infer_Tmap_from_multitime_clones_private(
                adata_temp,
                smooth_array=smooth_array,
                neighbor_N=CoSpar_KNN,
                sparsity_threshold=sparsity_threshold,
                intraclone_threshold=intraclone_threshold,
                normalization_mode=normalization_mode,
                save_subset=save_subset,
                use_full_Smatrix=use_full_Smatrix,
                trunca_threshold=trunca_threshold,
                compute_new_Smatrix=compute_new,
                max_iter_N=max_iter_N,
                epsilon_converge=epsilon_converge,
            )

            temp_id_t1 = np.nonzero(time_info == yy)[0]
            sp_id_t1 = hf.converting_id_from_fullSpace_to_subSpace(
                temp_id_t1, Tmap_cell_id_t1
            )[0]

            transition_map[sp_id_t1, :] = adata_temp.uns["transition_map"].A
            intraclone_transition_map[sp_id_t1, :] = adata_temp.uns[
                "intraclone_transition_map"
            ].A

            if "Smatrix" in adata_temp.uns.keys():
                adata_temp.uns.pop("Smatrix")

        adata.uns["transition_map"] = ssp.csr_matrix(transition_map)
        adata.uns["intraclone_transition_map"] = ssp.csr_matrix(
            intraclone_transition_map
        )

        logg.info(f"-----------Total used time: {time.time()-t0} s ------------")
        return adata


def infer_intraclone_Tmap(adata, intraclone_threshold=0.05, normalization_mode=1):
    """
    Infer intra-clone transition map.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` object
        Should be prepared by :func:`cospar.tmap._utils.select_time_points`
    intraclone_threshold: `float`, optional (default: 0.05)
        The threshold to remove noises in the demultiplexed (un-smoothed) map,
        in the range [0,1]
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.

    Returns
    -------
    None. Update/generate adata.uns['intraclone_transition_map']

    """

    ########## extract data
    if "transition_map" not in adata.uns.keys():
        logg.error(
            "Please run ---- CS.tmap.infer_Tmap_from_multitime_clones ---- first"
        )

    else:

        clone_annot = adata.obsm["X_clone"]

        multiTime_cell_id_t1 = [adata.uns["Tmap_cell_id_t1"]]
        multiTime_cell_id_t2 = [adata.uns["Tmap_cell_id_t2"]]
        proportion = adata.uns["proportion"]

        transition_map = adata.uns["transition_map"]

        X_clone = clone_annot.copy()
        if not ssp.issparse(X_clone):
            X_clone = ssp.csr_matrix(X_clone)

        demultiplexed_map = tmap_core.refine_Tmap_through_cospar_noSmooth(
            multiTime_cell_id_t1,
            multiTime_cell_id_t2,
            proportion,
            transition_map,
            X_clone,
            sparsity_threshold=intraclone_threshold,
            normalization_mode=normalization_mode,
        )

        adata.uns["intraclone_transition_map"] = ssp.csr_matrix(demultiplexed_map)


def infer_Tmap_from_one_time_clones(
    adata_orig,
    initial_time_points=None,
    later_time_point=None,
    initialize_method="OT",
    OT_epsilon=0.02,
    OT_dis_KNN=5,
    OT_cost="SPD",
    HighVar_gene_pctl=85,
    padding_X_clone=False,
    normalization_mode=1,
    sparsity_threshold=0.2,
    CoSpar_KNN=20,
    use_full_Smatrix=True,
    smooth_array=[15, 10, 5],
    trunca_threshold=[0.001, 0.01],
    compute_new=False,
    max_iter_N=[1, 5],
    epsilon_converge=[0.05, 0.05],
    use_fixed_clonesize_t1=False,
    sort_clone=1,
    save_subset=True,
    use_existing_KNN_graph=False,
):
    """
    Infer transition map from clones with a single time point

    We jointly infer a transition map and the initial clonal observation
    through iteration. The inferred map is between each of the initial
    time points ['day_1','day_2',...,] and the time point with clonal
    observation. We initialize the transition map by either the OT
    method or HighVar method.

    **Summary**

    * Parameters relevant for cell state selection:  initial_time_points,
      later_time_point.

    * Initialization methods:

        * 'OT': optional transport based method. Key parameters: `OT_epsilon, OT_dis_KNN`.
          See :func:`.infer_Tmap_from_optimal_transport`.

        * 'HighVar': a customized approach, assuming that cells similar in gene
          expression across time points share clonal origin. Key parameter: `HighVar_gene_pctl`.
          See :func:`.infer_Tmap_from_HighVar`.

    * Key parameters relevant for joint optimization itself (which relies on coherent sparse optimization):
      `smooth_array, CoSpar_KNN, sparsity_threshold`. See :func:`.refine_Tmap_through_joint_optimization`.


    Parameters
    ----------
    adata_orig: :class:`~anndata.AnnData` object
        It is assumed to be preprocessed and has multiple time points.
    initial_time_points: `list`, optional (default, all time points)
        List of initial time points to be included for the transition map.
        Like ['day_1','day_2']. Entries consistent with adata.obs['time_info'].
    later_time_point: `str`, optional (default, the last time point)
        The time point with clonal observation. Its value should be
        consistent with adata.obs['time_info'].
    initialize_method: `str`, optional (default 'OT')
        Method to initialize the transition map from state information.
        Choice: {'OT', 'HighVar'}.
    OT_epsilon: `float`, optional (default: 0.02)
        The entropic regularization, >0. A larger value increases
        uncertainty of the transition. Relevant when `initialize_method='OT'`.
    OT_dis_KNN: `int`, optional (default: 5)
        Number of nearest neighbors to construct the KNN graph for
        computing the shortest path distance. Relevant when `initialize_method='OT'`.
    OT_cost: `str`, optional (default: `SPD`), options {'GED','SPD'}
        The cost metric. We provide gene expression distance (GED), and also
        shortest path distance (SPD). GED is much faster, but SPD is more accurate.
        However, cospar is robust to the initialization.
    HighVar_gene_pctl: `int`, optional (default: 85)
        Percentile threshold to select highly variable genes to construct pseudo-clones.
        A higher value selects more variable genes. Range: [0,100].
        Relevant when `initialize_method='HighVar'`.
    padding_X_clone: `bool`, optional (default: False)
        If true, select cells at the `later_time_point` yet without any clonal label, and
        generate a unique clonal label for each of them. This adds artificial clonal data.
        However, it will make the best use of the state information, especially when there
        are very few clonal barcodes in the data.
    normalization_mode: `int`, optional (default: 1)
        Normalization method. Choice: [0,1].
        0, single-cell normalization; 1, Clone normalization. The clonal
        normalization suppresses the contribution of large
        clones, and is much more robust.
    smooth_array: `list`, optional (default: [15,10,5])
        List of smooth rounds at initial runs of iteration.
        Suppose that it has a length N. For iteration n<N, the n-th entry of
        smooth_array determines the kernel exponent to build the S matrix at the n-th
        iteration. When n>N, we use the last entry of smooth_array to compute
        the S matrix. We recommend starting with more smoothing depth and gradually
        reduce the depth, as inspired by simulated annealing. Data with higher
        clonal dispersion should start with higher smoothing depth. The final depth should
        depend on the manifold itself. For fewer cells, it results in a small KNN graph,
        and a small final depth should be used. We recommend to use a number at
        the multiple of 5 for computational efficiency i.e.,
        smooth_array=[20, 15, 10, 5], or [20,15,10]
    max_iter_N: `list`, optional (default: [1,5])
        A list for maximum iterations for the Joint optimization and CoSpar core function, respectively.
    epsilon_converge: `list`, optional (default: [0.05,0.05])
        A list of convergence threshold for the Joint optimization and CoSpar core function, respectively.
        The convergence threshold is for the change of map correlations between consecutive iterations.
        For CoSpar core function, this convergence test is activated only when CoSpar has iterated for 3 times.
    CoSpar_KNN: `int`, optional (default: 20)
        The number of neighbors for KNN graph used for computing the similarity matrix.
    trunca_threshold: `list`, optional (default: [0.001,0.01])
        Threshold to reset entries of a matrix to zero. The first entry is for
        Similarity matrix; the second entry is for the Tmap.
        This is only for computational and storage efficiency.
    sparsity_threshold: `float`, optional (default: 0.1)
        The relative threshold to remove noises in the updated transition map,
        in the range [0,1].
    save_subset: `bool`, optional (default: True)
        If true, save only Smatrix at smooth round [5,10,15,...];
        Otherwise, save Smatrix at each round.
    use_full_Smatrix: `bool`, optional (default: True)
        If true, extract the relevant Smatrix from the full Smatrix defined by all cells.
        This tends to be more accurate. The package is optimized around this choice.
    use_fixed_clonesize_t1: `bool`, optional (default: False)
        If true, fix the number of initial states as the same for all clones
    sort_clone: `int`, optional (default: 1)
        The order to infer initial states for each clone: {1,-1,others}.
        1, sort clones by size from small to large;
        -1, sort clones by size from large to small;
        others, do not sort.
    compute_new: `bool`, optional (default: False)
        If True, compute everything (ShortestPathDis, OT_map, etc.) from scratch,
        whether it was computed and saved before or not. Regarding the Smatrix, it is
        recomputed only when `use_full_Smatrix=False`.
    use_existing_KNN_graph: `bool`, optional (default: False)
        If true and adata.obsp['connectivities'], use the existing knn graph
        to compute the shortest-path distance. Revelant if initialize_method='OT'.
        This overrides all other relevant parameters for building shortest-path distance.

    Returns
    -------
    adata: :class:`~anndata.AnnData` object
        Update adata.obsm['X_clone'] and adata.uns['transition_map'],
        as well as adata.uns['OT_transition_map'] or
        adata.uns['HighVar_transition_map'], depending on the initialization.
        adata_orig.obsm['X_clone'] remains the same.
    """

    t0 = time.time()
    hf.check_available_clonal_info(adata_orig)
    clonal_time_points_0 = np.array(adata_orig.uns["clonal_time_points"])
    time_ordering = adata_orig.uns["time_ordering"]
    if len(clonal_time_points_0) == 0:
        raise ValueError(
            "No clonal time points available for this dataset. Please run cs.tmap.infer_Tmap_from_state_info_alone."
        )

    if later_time_point is None:
        sel_idx_temp = np.in1d(time_ordering, clonal_time_points_0)
        later_time_point = time_ordering[sel_idx_temp][-1]

    if type(later_time_point) == list:
        later_time_point = later_time_point[0]

    # use the last clonal later time point

    if initial_time_points is None:
        sel_id_temp = np.nonzero(np.in1d(time_ordering, [later_time_point]))[0][0]
        initial_time_points = time_ordering[:sel_id_temp]

    sel_idx_temp = np.in1d(time_ordering, initial_time_points)
    initial_time_points = list(time_ordering[sel_idx_temp])
    if later_time_point in initial_time_points:
        logg.warn(f"remove {later_time_point} from initial_time_points")
        initial_time_points.remove(later_time_point)

    hf.check_input_parameters(
        adata_orig,
        later_time_point=later_time_point,
        initial_time_points=initial_time_points,
        smooth_array=smooth_array,
        save_subset=save_subset,
    )

    if initialize_method not in ["OT", "HighVar"]:
        logg.warn(
            "initialize_method not among ['OT','HighVar']. Use initialize_method='OT'"
        )
        initialize_method = "OT"

    if OT_cost not in ["GED", "SPD"]:
        logg.warn("OT_cost not among ['GED','SPD']. Use OT_cost='SPD'")
        OT_cost = "SPD"

    sp_idx = np.zeros(adata_orig.shape[0], dtype=bool)
    time_info_orig = np.array(adata_orig.obs["time_info"])
    all_time_points = list(initial_time_points) + [later_time_point]

    label = "t"
    for xx in all_time_points:
        id_array = np.nonzero(time_info_orig == xx)[0]
        sp_idx[id_array] = True
        label = label + "*" + str(xx)

    adata = adata_orig[sp_idx]

    clone_annot_orig = adata_orig.obsm["X_clone"].copy()
    data_des_orig = adata_orig.uns["data_des"][0]
    data_des_0 = adata_orig.uns["data_des"][-1]
    data_des = data_des_0 + f"_OneTimeClone_{label}"
    adata.uns["data_des"] = [data_des_orig, data_des]

    time_info = np.array(adata.obs["time_info"])
    time_index_t2 = time_info == later_time_point
    time_index_t1 = ~time_index_t2

    ## set cells without a clone ID to have a unique clone ID
    if padding_X_clone:
        logg.info("Generate a unique clonal label for each clonally unlabeled cell.")
        time_index_t2_orig = time_info_orig == later_time_point
        zero_clone_idx = clone_annot_orig[time_index_t2_orig].sum(1).A.flatten() == 0
        clone_annot_t2_padding = np.diag(np.ones(np.sum(zero_clone_idx)))
        non_zero_clones_idx = (
            clone_annot_orig[time_index_t2_orig].sum(0).A.flatten() > 0
        )
        M0 = np.sum(non_zero_clones_idx)
        M1 = clone_annot_t2_padding.shape[1]
        clone_annot_new = np.zeros((clone_annot_orig.shape[0], M0 + M1))
        clone_annot_new[:, :M0] = clone_annot_orig[:, non_zero_clones_idx].A
        sp_id_t2 = np.nonzero(time_index_t2_orig)[0]
        clone_annot_new[sp_id_t2[zero_clone_idx], M0:] = clone_annot_t2_padding
    else:
        clone_annot_new = clone_annot_orig

    # remove clones without a cell at t2
    valid_clone_id = np.nonzero(
        clone_annot_new[time_info_orig == later_time_point].sum(0).A.flatten() > 0
    )[0]
    X_clone_temp = clone_annot_new[:, valid_clone_id]
    adata_orig.obsm["X_clone"] = ssp.csr_matrix(X_clone_temp)

    #### used for similarity matrix generation
    Tmap_cell_id_t1 = np.nonzero(time_index_t1)[0]
    Tmap_cell_id_t2 = np.nonzero(time_index_t2)[0]
    adata.uns["Tmap_cell_id_t1"] = Tmap_cell_id_t1
    adata.uns["Tmap_cell_id_t2"] = Tmap_cell_id_t2
    adata.uns["clonal_cell_id_t1"] = Tmap_cell_id_t1
    adata.uns["clonal_cell_id_t2"] = Tmap_cell_id_t2
    adata.uns["sp_idx"] = sp_idx
    data_path = settings.data_path

    transition_map = np.zeros((len(Tmap_cell_id_t1), len(Tmap_cell_id_t2)))
    ini_transition_map = np.zeros((len(Tmap_cell_id_t1), len(Tmap_cell_id_t2)))
    X_clone_updated = adata_orig.obsm["X_clone"][
        sp_idx
    ].A  # this does not work well if there are empty clones to begin with

    logg.info(
        "--------Infer transition map between initial time points and the later time one-------"
    )
    for yy in initial_time_points:
        logg.info(f"--------Current initial time point: {yy}--------")

        adata_temp = infer_Tmap_from_one_time_clones_twoTime(
            adata_orig,
            selected_two_time_points=[yy, later_time_point],
            initialize_method=initialize_method,
            OT_epsilon=OT_epsilon,
            OT_dis_KNN=OT_dis_KNN,
            OT_cost=OT_cost,
            HighVar_gene_pctl=HighVar_gene_pctl,
            normalization_mode=normalization_mode,
            sparsity_threshold=sparsity_threshold,
            CoSpar_KNN=CoSpar_KNN,
            use_full_Smatrix=use_full_Smatrix,
            smooth_array=smooth_array,
            trunca_threshold=trunca_threshold,
            compute_new=compute_new,
            use_fixed_clonesize_t1=use_fixed_clonesize_t1,
            sort_clone=sort_clone,
            save_subset=save_subset,
            use_existing_KNN_graph=use_existing_KNN_graph,
            max_iter_N=max_iter_N,
            epsilon_converge=epsilon_converge,
        )

        temp_id_t1 = np.nonzero(time_info == yy)[0]
        sp_id_t1 = hf.converting_id_from_fullSpace_to_subSpace(
            temp_id_t1, Tmap_cell_id_t1
        )[0]

        transition_map_temp = adata_temp.uns["transition_map"].A
        transition_map[sp_id_t1, :] = transition_map_temp

        if initialize_method == "OT":
            transition_map_ini_temp = adata_temp.uns["OT_transition_map"]
        else:
            transition_map_ini_temp = adata_temp.uns["HighVar_transition_map"]

        ini_transition_map[sp_id_t1, :] = transition_map_ini_temp.A

        # Update clonal prediction. This does not work well if there are empty clones to begin with
        time_info_idx = np.array(adata_temp.obs["time_info"]) == yy
        X_clone_updated[temp_id_t1, :] = adata_temp.obsm["X_clone"][time_info_idx].A

    adata.uns["transition_map"] = ssp.csr_matrix(transition_map)
    adata.obsm["X_clone"] = ssp.csr_matrix(X_clone_updated)

    if initialize_method == "OT":
        adata.uns["OT_transition_map"] = ssp.csr_matrix(ini_transition_map)
    else:
        adata.uns["HighVar_transition_map"] = ssp.csr_matrix(ini_transition_map)

    adata_orig.obsm["X_clone"] = clone_annot_orig  # reset to the original clonal matrix
    logg.info(f"-----------Total used time: {time.time()-t0} s ------------")
    return adata


# updated version: v1, we initialize the X_clone as isolated cells
def infer_Tmap_from_state_info_alone(
    adata_orig,
    initial_time_points=None,
    later_time_point=None,
    initialize_method="OT",
    OT_epsilon=0.02,
    OT_dis_KNN=5,
    OT_cost="SPD",
    HighVar_gene_pctl=85,
    normalization_mode=1,
    sparsity_threshold=0.2,
    CoSpar_KNN=20,
    use_full_Smatrix=True,
    smooth_array=[15, 10, 5],
    trunca_threshold=[0.001, 0.01],
    compute_new=False,
    max_iter_N=[1, 5],
    epsilon_converge=[0.05, 0.05],
    use_fixed_clonesize_t1=False,
    sort_clone=1,
    save_subset=True,
    use_existing_KNN_graph=False,
):
    """
    Infer transition map from state information alone.

    After initializing the clonal matrix as such that each cell has a unique barcode,
    it runs :func:`.infer_Tmap_from_one_time_clones` to infer the transition map. Please see :func:`.infer_Tmap_from_one_time_clones` for the meaning of each parameter.

    Returns
    -------
    adata will include both the inferred transition map, and also the updated X_clone matrix.
    The input, adata_orig, will maintain the original X_clone matrix.
    """

    ##--------------- check input parameters
    if "data_des" not in adata_orig.uns.keys():
        adata_orig.uns["data_des"] = ["cospar"]
    logg.info(
        "Step I: Generate pseudo clones where each cell has a unique barcode-----"
    )

    if type(later_time_point) == list:
        later_time_point = later_time_point[0]

    hf.update_time_ordering(adata_orig, mode="auto")
    time_ordering = adata_orig.uns["time_ordering"]

    # use the last time point
    if later_time_point is None:
        later_time_point = time_ordering[-1]

    if initial_time_points is None:
        # use the time points preceding the last one.
        sel_id_temp = np.nonzero(np.in1d(time_ordering, [later_time_point]))[0][0]
        initial_time_points = time_ordering[:sel_id_temp]
    else:
        # re-order time points. This also gets rid of invalid time points
        sel_idx_temp = np.in1d(time_ordering, initial_time_points)
        if np.sum(sel_idx_temp) > 0:
            initial_time_points = time_ordering[sel_idx_temp]
        else:
            raise ValueError(
                f"The 'initial_time_points' are not valid. Please select from {time_ordering}"
            )

    ##--------------- use the artifical clonal matrix
    X_clone_0 = adata_orig.obsm["X_clone"].copy()
    # adata_orig.obsm['X_clone_old']=adata_orig.obsm['X_clone'].copy()
    X_clone = np.diag(np.ones(adata_orig.shape[0]))
    adata_orig.obsm["X_clone"] = ssp.csr_matrix(X_clone)

    logg.info("Step II: Perform joint optimization-----")
    adata = infer_Tmap_from_one_time_clones(
        adata_orig,
        initial_time_points=initial_time_points,
        later_time_point=later_time_point,
        initialize_method=initialize_method,
        OT_epsilon=OT_epsilon,
        OT_dis_KNN=OT_dis_KNN,
        OT_cost=OT_cost,
        HighVar_gene_pctl=HighVar_gene_pctl,
        normalization_mode=normalization_mode,
        sparsity_threshold=sparsity_threshold,
        CoSpar_KNN=CoSpar_KNN,
        use_full_Smatrix=use_full_Smatrix,
        smooth_array=smooth_array,
        trunca_threshold=trunca_threshold,
        compute_new=compute_new,
        max_iter_N=max_iter_N,
        epsilon_converge=epsilon_converge,
        use_fixed_clonesize_t1=use_fixed_clonesize_t1,
        sort_clone=sort_clone,
        save_subset=save_subset,
        use_existing_KNN_graph=use_existing_KNN_graph,
    )

    # only restore the original X_clone information to adata_orig.
    # adata will carry the new structure
    adata_orig.obsm["X_clone"] = X_clone_0

    time_info_orig = np.array(adata_orig.obs["time_info"])
    all_time_points = list(initial_time_points) + [later_time_point]
    label = "t"
    for xx in all_time_points:
        id_array = np.nonzero(time_info_orig == xx)[0]
        label = label + "*" + str(xx)

    data_des_orig = adata_orig.uns["data_des"][0]
    data_des_0 = adata_orig.uns["data_des"][-1]
    data_des = data_des_0 + f"_StateInfo_{label}"
    adata.uns["data_des"] = [data_des_orig, data_des]

    return adata


def infer_Tmap_from_one_time_clones_twoTime(
    adata_orig,
    selected_two_time_points=["1", "2"],
    initialize_method="OT",
    OT_epsilon=0.02,
    OT_dis_KNN=5,
    OT_cost="SPD",
    HighVar_gene_pctl=80,
    normalization_mode=1,
    sparsity_threshold=0.2,
    CoSpar_KNN=20,
    use_full_Smatrix=True,
    smooth_array=[15, 10, 5],
    max_iter_N=[1, 5],
    epsilon_converge=[0.05, 0.05],
    trunca_threshold=[0.001, 0.01],
    compute_new=True,
    use_fixed_clonesize_t1=False,
    sort_clone=1,
    save_subset=True,
    joint_optimization=True,
    use_existing_KNN_graph=False,
):
    """
    Infer transition map from clones with a single time point

    It is the same as :func:`.infer_Tmap_from_one_time_clones`, except that
    it assumes that the input adata_orig has only two time points.

    joint_optimization: `bool`, optional (default: True).
    """

    time_info_orig = np.array(adata_orig.obs["time_info"])
    sort_time_point = np.sort(list(set(time_info_orig)))
    N_valid_time = np.sum(np.in1d(sort_time_point, selected_two_time_points))
    if N_valid_time != 2:
        logg.error(f"Must select only two time points among the list {sort_time_point}")
        # The second time point in this list (not necessarily later time point) is assumed to have clonal data.")
    else:
        ####################################

        logg.info("Step 0: Pre-processing and sub-sampling cells-------")
        # select cells from the two time points, and sub-sampling, create the new adata object with these cell states
        sp_idx = (time_info_orig == selected_two_time_points[0]) | (
            time_info_orig == selected_two_time_points[1]
        )

        adata = adata_orig[sp_idx]
        data_des_0 = adata_orig.uns["data_des"][-1]
        data_des_orig = adata_orig.uns["data_des"][0]
        data_des = (
            data_des_0
            + f"_t*{selected_two_time_points[0]}*{selected_two_time_points[1]}"
        )
        adata.uns["data_des"] = [data_des_orig, data_des]

        time_info = np.array(adata.obs["time_info"])
        time_index_t1 = time_info == selected_two_time_points[0]
        time_index_t2 = time_info == selected_two_time_points[1]

        #### used for similarity matrix generation
        Tmap_cell_id_t1 = np.nonzero(time_index_t1)[0]
        Tmap_cell_id_t2 = np.nonzero(time_index_t2)[0]
        adata.uns["Tmap_cell_id_t1"] = Tmap_cell_id_t1
        adata.uns["Tmap_cell_id_t2"] = Tmap_cell_id_t2
        adata.uns["clonal_cell_id_t1"] = Tmap_cell_id_t1
        adata.uns["clonal_cell_id_t2"] = Tmap_cell_id_t2
        adata.uns["sp_idx"] = sp_idx
        data_path = settings.data_path

        ###############################
        # prepare the similarity matrix with all state info, all subsequent similarity will be down-sampled from this one.
        if use_full_Smatrix and (joint_optimization or (initialize_method != "OT")):

            temp_str = "0" + str(trunca_threshold[0])[2:]
            round_of_smooth = np.max(smooth_array)
            data_des = adata_orig.uns["data_des"][0]
            similarity_file_name = os.path.join(
                data_path,
                f"{data_des}_Similarity_matrix_with_all_cell_states_kNN{CoSpar_KNN}_Truncate{temp_str}",
            )
            if not (
                os.path.exists(similarity_file_name + f"_SM{round_of_smooth}.npz")
                and (not compute_new)
            ):
                similarity_matrix_full = tmap_util.generate_similarity_matrix(
                    adata_orig,
                    similarity_file_name,
                    round_of_smooth=round_of_smooth,
                    neighbor_N=CoSpar_KNN,
                    truncation_threshold=trunca_threshold[0],
                    save_subset=save_subset,
                    compute_new_Smatrix=compute_new,
                )

        if initialize_method == "OT":

            # logg.info("----------------")
            logg.info("Step 1: Use OT method for initialization-------")

            tmap_core.infer_Tmap_from_optimal_transport(
                adata,
                OT_epsilon=OT_epsilon,
                OT_cost=OT_cost,
                OT_dis_KNN=OT_dis_KNN,
                compute_new=compute_new,
                use_existing_KNN_graph=use_existing_KNN_graph,
            )

            OT_transition_map = adata.uns["OT_transition_map"]
            initialized_map = OT_transition_map

        else:

            # logg.info("----------------")
            logg.info("Step 1: Use the HighVar method for initialization-------")

            t = time.time()
            tmap_core.infer_Tmap_from_HighVar(
                adata,
                min_counts=3,
                min_cells=3,
                min_gene_vscore_pctl=HighVar_gene_pctl,
                sparsity_threshold=sparsity_threshold,
                neighbor_N=CoSpar_KNN,
                normalization_mode=normalization_mode,
                use_full_Smatrix=use_full_Smatrix,
                smooth_array=smooth_array,
                trunca_threshold=trunca_threshold,
                compute_new_Smatrix=compute_new,
                max_iter_N=max_iter_N[1],
                epsilon_converge=epsilon_converge[1],
            )

            HighVar_transition_map = adata.uns["HighVar_transition_map"]
            initialized_map = HighVar_transition_map
            logg.info(
                f"Finishing initialization using HighVar, used time {time.time()-t}"
            )

        if joint_optimization:
            ########### Jointly optimize the transition map and the initial clonal states
            if selected_two_time_points[1] in adata_orig.uns["clonal_time_points"]:

                # logg.info("----------------")
                logg.info(
                    "Step 2: Jointly optimize the transition map and the initial clonal states-------"
                )

                t = time.time()

                tmap_core.refine_Tmap_through_joint_optimization(
                    adata,
                    initialized_map,
                    normalization_mode=normalization_mode,
                    sparsity_threshold=sparsity_threshold,
                    CoSpar_KNN=CoSpar_KNN,
                    use_full_Smatrix=use_full_Smatrix,
                    smooth_array=smooth_array,
                    max_iter_N=max_iter_N,
                    epsilon_converge=epsilon_converge,
                    trunca_threshold=trunca_threshold,
                    compute_new=compute_new,
                    use_fixed_clonesize_t1=use_fixed_clonesize_t1,
                    sort_clone=sort_clone,
                    save_subset=save_subset,
                )

                logg.info(f"Finishing Joint Optimization, used time {time.time()-t}")
            else:
                logg.warn(
                    "No clonal information available. Skip the joint optimization of clone and scRNAseq data"
                )

        if "Smatrix" in adata.uns.keys():
            adata.uns.pop("Smatrix")
        return adata


def infer_Tmap_from_clonal_info_alone_private(
    adata_orig, method="naive", clonal_time_points=None, selected_fates=None
):
    """
    Compute transition map using only the lineage information.

    Here, we compute the transition map between neighboring time points.

    We simply average transitions across all clones (or selected clones when method='Weinreb'),
    assuming that the intra-clone transition is uniform within the same clone.

    Parameters
    ----------
    adata_orig: :class:`~anndata.AnnData` object
    method: `str`, optional (default: 'naive')
        Method used to compute the transition map. Choice: {'naive',
        'weinreb'}. For the naive method, we simply average transitions
        across all clones, assuming that the intra-clone transitions are
        uniform within the same clone. For the 'weinreb' method, we first
        find uni-potent clones, then compute the transition map by simply
        averaging across all clonal transitions as the naive method.
    selected_fates: `list`, optional (default: all selected)
        List of targeted fate clusters to define uni-potent clones for the
        weinreb method, which are used to compute the transition map.
    clonal_time_points: `list` of `str`, optional (default: all time points)
        List of time points to be included for analysis.
        We assume that each selected time point has clonal measurements.
    later_time_points: `list`, optional (default: None)
        If specified, the function will produce a map T between these early
        time points among `clonal_time_points` and the `later_time_point`.
        If not specified, it produces a map T between neighboring time points.

    Returns
    -------
    adata: :class:`~anndata.AnnData` object
        The transition map is stored at adata.uns['clonal_transition_map']
    """

    adata_1 = tmap_util.select_time_points(
        adata_orig, time_point=clonal_time_points, extend_Tmap_space=True
    )
    if method not in ["naive", "weinreb"]:
        logg.warn("method not in ['naive','weinreb']; set to be 'weinreb'")
        method = "weinreb"

    cell_id_t2_all = adata_1.uns["Tmap_cell_id_t2"]
    cell_id_t1_all = adata_1.uns["Tmap_cell_id_t1"]

    T_map = np.zeros((len(cell_id_t1_all), len(cell_id_t2_all)))
    clone_annot = adata_1.obsm["X_clone"]

    N_points = len(adata_1.uns["multiTime_cell_id_t1"])
    for k in range(N_points):

        cell_id_t1_temp = adata_1.uns["multiTime_cell_id_t1"][k]
        cell_id_t2_temp = adata_1.uns["multiTime_cell_id_t2"][k]
        if method == "naive":
            logg.info("Use all clones (naive method)")
            T_map_temp = clone_annot[cell_id_t1_temp] * clone_annot[cell_id_t2_temp].T

        else:
            logg.info("Use only uni-potent clones (weinreb et al., 2020)")
            state_annote = np.array(adata_1.obs["state_info"])
            if selected_fates == None:
                selected_fates = list(set(state_annote))
            potential_vector_clone, fate_entropy_clone = tl.compute_state_potential(
                clone_annot[cell_id_t2_temp].T,
                state_annote[cell_id_t2_temp],
                selected_fates,
                fate_count=True,
            )

            sel_unipotent_clone_id = np.array(
                list(set(np.nonzero(fate_entropy_clone == 1)[0]))
            )
            clone_annot_unipotent = clone_annot[:, sel_unipotent_clone_id]
            T_map_temp = (
                clone_annot_unipotent[cell_id_t1_temp]
                * clone_annot_unipotent[cell_id_t2_temp].T
            )
            logg.info(
                f"Used uni-potent clone fraction {len(sel_unipotent_clone_id)/clone_annot.shape[1]}"
            )

        idx_t1 = np.nonzero(np.in1d(cell_id_t1_all, cell_id_t1_temp))[0]
        idx_t2 = np.nonzero(np.in1d(cell_id_t2_all, cell_id_t2_temp))[0]
        idx_t1_temp = np.nonzero(np.in1d(cell_id_t1_temp, cell_id_t1_all))[0]
        idx_t2_temp = np.nonzero(np.in1d(cell_id_t2_temp, cell_id_t2_all))[0]
        T_map[idx_t1[:, np.newaxis], idx_t2] = T_map_temp[idx_t1_temp][:, idx_t2_temp].A

    T_map = T_map.astype(int)
    adata_1.uns["clonal_transition_map"] = ssp.csr_matrix(T_map)
    return adata_1


# the v2 version, it is the same format as infer_Tmap_from_multiTime_clones.
# We return a new adata object that will throw away existing annotations in uns.
def infer_Tmap_from_clonal_info_alone(
    adata_orig,
    method="naive",
    clonal_time_points=None,
    later_time_point=None,
    selected_fates=None,
):
    """
    Compute transition map using only the lineage information.

    As in :func:`.infer_Tmap_from_multitime_clones`, we provide two modes of inference:

    * If `later_time_point=None`, the inferred map allows transitions
      between neighboring time points. For example, if
      clonal_time_points=['day1','day2','day3'], then it computes transitions
      for pairs (day1, day2) and (day2, day3), but not (day1, day3).

    * If `later_time_point` is specified, the function produces a map
      between earlier time points and this later time point. For example, if
      `later_time_point='day3`, the map allows transitions for pairs (day1, day3)
      and (day2, day3), but not (day1,day2).

    Parameters
    ----------
    adata_orig: :class:`~anndata.AnnData` object
    method: `str`, optional (default: 'naive')
        Method used to compute the transition map. Choice: {'naive',
        'weinreb'}. For the naive method, we simply average transitions
        across all clones, assuming that the intra-clone transitions are
        uniform within the same clone. For the 'weinreb' method, we first
        find uni-potent clones, then compute the transition map by simply
        averaging across all clonal transitions as the naive method.
    selected_fates: `list`, optional (default: all selected)
        List of targeted fate clusters to define uni-potent clones for the
        weinreb method, which are used to compute the transition map.
    clonal_time_points: `list` of `str`, optional (default: all time points)
        List of time points to be included for analysis.
        We assume that each selected time point has clonal measurements.
    later_time_points: `list`, optional (default: None)
        If specified, the function will produce a map T between these early
        time points among `clonal_time_points` and the `later_time_point`.
        If not specified, it produces a map T between neighboring time points.

    Returns
    -------
    adata: :class:`~anndata.AnnData` object
        The transition map is stored at adata.uns['clonal_transition_map']
    """

    hf.check_available_clonal_info(adata_orig)
    clonal_time_points_0 = np.array(adata_orig.uns["clonal_time_points"])
    if len(clonal_time_points_0) < 2:
        raise ValueError("There are no multi-time clones. Abort the inference.")

    if clonal_time_points is None:
        clonal_time_points = clonal_time_points_0

    if type(later_time_point) == list:
        later_time_point = later_time_point[0]

    if later_time_point is not None:
        clonal_time_points = list(clonal_time_points) + [later_time_point]
        clonal_time_points = list(set(clonal_time_points))

    hf.check_input_parameters(
        adata_orig,
        later_time_point=later_time_point,
        clonal_time_points=clonal_time_points,
    )
    # order the clonal time points
    time_ordering = adata_orig.uns["time_ordering"]
    sel_idx_temp = np.in1d(time_ordering, clonal_time_points)
    clonal_time_points = time_ordering[sel_idx_temp]

    if later_time_point is None:
        logg.info("Infer transition map between neighboring time points.")
        adata = infer_Tmap_from_clonal_info_alone_private(
            adata_orig,
            method=method,
            clonal_time_points=clonal_time_points,
            selected_fates=selected_fates,
        )

        return adata
    else:
        logg.info(
            f"Infer transition map between initial time points and the later time point."
        )
        # compute transition map between initial time points and the later time point
        sel_id = np.nonzero(np.in1d(clonal_time_points, later_time_point))[0][0]
        initial_time_points = clonal_time_points[:sel_id]

        time_info_orig = np.array(adata_orig.obs["time_info"])
        sp_idx = np.zeros(adata_orig.shape[0], dtype=bool)
        all_time_points = list(initial_time_points) + [later_time_point]
        label = "t"
        for xx in all_time_points:
            id_array = np.nonzero(time_info_orig == xx)[0]
            sp_idx[id_array] = True
            label = label + "*" + str(xx)

        adata = adata_orig[sp_idx]
        data_des_orig = adata_orig.uns["data_des"][0]
        data_des_0 = adata_orig.uns["data_des"][-1]
        data_des = data_des_0 + f"_ClonalMap_Later_{label}"
        adata_orig.uns["data_des"] = [data_des_orig, data_des]

        time_info = np.array(adata_orig.obs["time_info"])
        time_index_t2 = time_info == later_time_point
        time_index_t1 = ~time_index_t2

        #### used for similarity matrix generation
        Tmap_cell_id_t1 = np.nonzero(time_index_t1)[0]
        Tmap_cell_id_t2 = np.nonzero(time_index_t2)[0]
        adata.uns["Tmap_cell_id_t1"] = Tmap_cell_id_t1
        adata.uns["Tmap_cell_id_t2"] = Tmap_cell_id_t2
        adata.uns["clonal_cell_id_t1"] = Tmap_cell_id_t1
        adata.uns["clonal_cell_id_t2"] = Tmap_cell_id_t2
        adata.uns["sp_idx"] = sp_idx
        data_path = settings.data_path

        transition_map = np.zeros((len(Tmap_cell_id_t1), len(Tmap_cell_id_t2)))

        # logg.info("------Infer transition map between initial time points and the later time one-------")
        for yy in initial_time_points:
            logg.info(f"--------Current initial time point: {yy}--------")

            # by default, we extend the state space to all cells at the given time point.
            adata_temp = infer_Tmap_from_clonal_info_alone_private(
                adata_orig,
                method=method,
                clonal_time_points=[yy, later_time_point],
                selected_fates=selected_fates,
            )

            temp_id_t1 = np.nonzero(time_info == yy)[0]
            sp_id_t1 = hf.converting_id_from_fullSpace_to_subSpace(
                temp_id_t1, Tmap_cell_id_t1
            )[0]

            # by default, we extend the state space to all cells at the given time point.
            # so we only need to care about t1.
            transition_map[sp_id_t1, :] = adata_temp.uns["clonal_transition_map"].A

        adata.uns["clonal_transition_map"] = ssp.csr_matrix(transition_map)

        return adata



"""KAILINss - dynamic inference by integrating transcriptome and lineage information"""

#计算稀疏矩阵matrix
#如果输入draw，就转变成非稀疏矩阵并开始画图

def kl_sparse_matrix_sample(
    sample_name,
    data_name,
    if_draw,
    current_folder_input,
    round_of_smooth,
    neighbor_N,
    beta,
    truncation_threshold,
    save_subset,
    use_existing_KNN_graph=False,
    compute_new_Smatrix=True,
    use_full_Smatrix = True
    ):

    global test_chosen_first_time_or_not #这个部分让计算matrix只运行一次。如果启用调试，则每个
    global loading_directory

    import os
    import anndata
    import cospar as cs
    import os
    import sys
    from LittleSnowFox.omic_functions import matrix                           
    import numpy as np
    import pandas as pd
    import scanpy as sc
    import scipy.sparse as ssp
    from cospar.tmap import _tmap_core as tmap_core
    from cospar.tmap import _utils as tmap_util
    from cospar import help_functions as hf
    from cospar import logging as logg
    from cospar import settings
    from cospar import tool as tl

    current_folder = current_folder_input

    if 1:
        sample_name = sample_name
        print("Sample: ",sample_name)
        current_dir = os.getcwd()
        print("Changing the working directory from:", current_dir)
        #eng.run()
        loading_directory = os.path.join(current_folder, sample_name)
        kl_data_menu = "data"
        data_directory = os.path.join(loading_directory, kl_data_menu)
    
    print() 
    print("---------------------------------------------------------------------------------------")
    print("Please put the data at: ",loading_directory,"\data")
    print("---------------------------------------------------------------------------------------")
    print()
    print("Current working directory is:", data_directory)

    # 构建data.h5ad文件的完整路径
    h5ad_file_path = os.path.join(data_directory, data_name)
    adata_using = anndata.read_h5ad(h5ad_file_path)
    import scipy.sparse as sp
    # 将adata.X转换为稀疏矩阵类型
    sparse_X = sp.csr_matrix(adata_using.X.astype(float))
    adata_using.X=sparse_X
    import scipy.sparse as sp


    if sample_name == "Reprogramming":
        distance_matrix = matrix.Reprogramming_distance_matrix(adata_using)
    if sample_name == "Hematopoiesis":
        distance_matrix = matrix.Hematopoiesis_distance_matrix(adata_using)
    else:
        distance_matrix = matrix.distance_matrix(
                            adata_using,
                            round_of_smooth,
                            neighbor_N,
                            beta,
                            truncation_threshold,
                            save_subset,
                            use_existing_KNN_graph=False,
                            compute_new_Smatrix=True,
                            use_full_Smatrix = True,
                            )

    if sp.issparse(distance_matrix):
        print("adata.X is changed as sparse matrix")
    else:
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")


    if if_draw == "draw":
       
        import seaborn as sns
        import matplotlib.pyplot as plt
        import numpy as np
        from scipy import sparse
        # 假设 similarity_matrix 是你的稀疏矩阵数据
        # 将稀疏矩阵转换为非稀疏矩阵
        dense_matrix = distance_matrix.toarray()
        if sp.issparse(adata_using.X):
            print("adata.X is changed as sparse matrix")
        else:
            print("adata.X is changed as dense matrix")
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues')
        # 添加标题和标签
        plt.title('Distance Matrix')
        plt.xlabel('Samples')
        plt.ylabel('Samples')
        # 显示热图
        plt.show()
        import seaborn as sns
        # 假设你已经有了一个稠密矩阵 dense_matrix
        # 设置热图的上下限
        vmin = 0  # 最小值
        vmax = 0.004  # 最大值
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues', vmin=vmin, vmax=vmax)
        # 显示图形
        plt.show()


    return adata_using,loading_directory,distance_matrix
    





#计算非稀疏matrix：
#如果输入draw，开始画图

def kl_dense_matrix_sample(
    sample_name,
    data_name,
    if_draw,
    current_folder_input,
    round_of_smooth=15,
    neighbor_N=20,
    beta=0.1,
    truncation_threshold=0.001,
    save_subset=True,
    use_existing_KNN_graph=False,
    compute_new_Smatrix=True,
    use_full_Smatrix = True,
    ):

    global test_chosen_first_time_or_not #这个部分让计算matrix只运行一次。如果启用调试，则每个
    global loading_directory

    import os
    import anndata
    import cospar as cs
    import os
    import sys
    from LittleSnowFox.omic_functions import matrix                           
    import numpy as np
    import pandas as pd
    import scanpy as sc
    import scipy.sparse as ssp
    from cospar.tmap import _tmap_core as tmap_core
    from cospar.tmap import _utils as tmap_util
    from cospar import help_functions as hf
    from cospar import logging as logg
    from cospar import settings
    from cospar import tool as tl

    current_folder = current_folder_input

    if 1:
        sample_name = sample_name
        print("Sample: ",sample_name)
        current_dir = os.getcwd()
        print("Changing the working directory from:", current_dir)
        #eng.run()
        loading_directory = os.path.join(current_folder, sample_name)
        kl_data_menu = "data"
        data_directory = os.path.join(loading_directory, kl_data_menu)
    
    print() 
    print("---------------------------------------------------------------------------------------")
    print("Please put the data at: ",loading_directory,"\data")
    print("---------------------------------------------------------------------------------------")
    print()
    print("Current working directory is:", data_directory)

    # 构建data.h5ad文件的完整路径
    h5ad_file_path = os.path.join(data_directory, data_name)
    adata_using = anndata.read_h5ad(h5ad_file_path)
    import scipy.sparse as sp
    # 将adata.X转换为稀疏矩阵类型
    sparse_X = sp.csr_matrix(adata_using.X.astype(float))
    adata_using.X=sparse_X
    import scipy.sparse as sp

    if sample_name == "Reprogramming":
        distance_matrix = matrix.Reprogramming_distance_matrix(adata_using)
    elif sample_name == "Hematopoiesis":
        distance_matrix = matrix.Hematopoiesis_distance_matrix(adata_using)
    else:
        distance_matrix = matrix.distance_matrix(
                            adata_using,
                            round_of_smooth,
                            neighbor_N,
                            beta,
                            truncation_threshold,
                            save_subset,
                            use_existing_KNN_graph=False,
                            compute_new_Smatrix=True,
                            use_full_Smatrix = True,
                            )

    dense_matrix = distance_matrix.toarray()

    if sp.issparse(dense_matrix):
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
    else:
        print("adata.X is not a sparse matrix")


    if if_draw == "draw":
       
        import seaborn as sns
        import matplotlib.pyplot as plt
        import numpy as np
        from scipy import sparse
        # 假设 similarity_matrix 是你的稀疏矩阵数据
        # 将稀疏矩阵转换为非稀疏矩阵
        
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues')
        # 添加标题和标签
        plt.title('Distance Matrix')
        plt.xlabel('Samples')
        plt.ylabel('Samples')
        # 显示热图
        plt.show()
        import seaborn as sns
        # 假设你已经有了一个稠密矩阵 dense_matrix
        # 设置热图的上下限
        vmin = 0  # 最小值
        vmax = 0.004  # 最大值
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues', vmin=vmin, vmax=vmax)
        # 显示图形
        plt.show()


    return adata_using,loading_directory,dense_matrix


#------------------------------------------------------------------------------------------------------------------------------
'''
def Reprogramming_distance_matrix(
    adata,
    round_of_smooth,
    neighbor_N,
    beta,
    truncation_threshold,
    save_subset,
    use_existing_KNN_graph=False,
    compute_new_Smatrix=True,
    use_full_Smatrix = True,
):
'''

def kl_sparse_matrix(
        sample_name,
        data_name,
        if_draw,
        current_folder_input,
        adata,
        round_of_smooth,
        neighbor_N,
        beta,
        truncation_threshold,
        save_subset,
        use_existing_KNN_graph=False,
        compute_new_Smatrix=True,
        use_full_Smatrix = True
    ):

    global test_chosen_first_time_or_not #这个部分让计算matrix只运行一次。如果启用调试，则每个
    global loading_directory

    import os
    import anndata
    import cospar as cs
    import os
    import sys
    from LittleSnowFox.omic_functions import matrix                           
    import numpy as np
    import pandas as pd
    import scanpy as sc
    import scipy.sparse as ssp
    from cospar.tmap import _tmap_core as tmap_core
    from cospar.tmap import _utils as tmap_util
    from cospar import help_functions as hf
    from cospar import logging as logg
    from cospar import settings
    from cospar import tool as tl

    current_folder = current_folder_input

    if 1:
        sample_name = sample_name
        print("Sample: ",sample_name)
        current_dir = os.getcwd()
        print("Changing the working directory from:", current_dir)
        #eng.run()
        loading_directory = os.path.join(current_folder, sample_name)
        kl_data_menu = "data"
        data_directory = os.path.join(loading_directory, kl_data_menu)
    
    print() 
    print("---------------------------------------------------------------------------------------")
    print("Please put the data at: ",loading_directory,"\data")
    print("---------------------------------------------------------------------------------------")
    print()
    print("Current working directory is:", data_directory)

    # 构建data.h5ad文件的完整路径
    h5ad_file_path = os.path.join(data_directory, data_name)
    adata_using = anndata.read_h5ad(h5ad_file_path)
    import scipy.sparse as sp
    # 将adata.X转换为稀疏矩阵类型
    sparse_X = sp.csr_matrix(adata_using.X.astype(float))
    adata_using.X=sparse_X
    import scipy.sparse as sp


    distance_matrix = matrix.distance_matrix(
        adata_using,
        round_of_smooth,
        neighbor_N,
        beta,
        truncation_threshold,
        save_subset,
        use_existing_KNN_graph=False,
        compute_new_Smatrix=True,
        use_full_Smatrix = True,
    )

    if sp.issparse(distance_matrix):
        print("adata.X is changed as sparse matrix")
    else:
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")
        print("!!!ERROR!!!: adata.X is not a sparse matrix")


    if if_draw == "draw":
       
        import seaborn as sns
        import matplotlib.pyplot as plt
        import numpy as np
        from scipy import sparse
        # 假设 similarity_matrix 是你的稀疏矩阵数据
        # 将稀疏矩阵转换为非稀疏矩阵
        dense_matrix = distance_matrix.toarray()
        if sp.issparse(adata_using.X):
            print("adata.X is changed as sparse matrix")
        else:
            print("adata.X is changed as dense matrix")
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues')
        # 添加标题和标签
        plt.title('Distance Matrix')
        plt.xlabel('Samples')
        plt.ylabel('Samples')
        # 显示热图
        plt.show()
        import seaborn as sns
        # 假设你已经有了一个稠密矩阵 dense_matrix
        # 设置热图的上下限
        vmin = 0  # 最小值
        vmax = 0.004  # 最大值
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues', vmin=vmin, vmax=vmax)
        # 显示图形
        plt.show()


    return adata_using,loading_directory,distance_matrix
    





#计算非稀疏matrix：
#如果输入draw，开始画图

def kl_dense_matrix(
        sample_name,
        data_name,
        if_draw,
        current_folder_input,
        round_of_smooth,
        neighbor_N,
        beta,
        truncation_threshold,
        save_subset,
        use_existing_KNN_graph=False,
        compute_new_Smatrix=True,
        use_full_Smatrix = True
    ):

    global test_chosen_first_time_or_not #这个部分让计算matrix只运行一次。如果启用调试，则每个
    global loading_directory

    import os
    import anndata
    import cospar as cs
    import os
    import sys
    from LittleSnowFox.omic_functions import matrix                           
    import numpy as np
    import pandas as pd
    import scanpy as sc
    import scipy.sparse as ssp
    from cospar.tmap import _tmap_core as tmap_core
    from cospar.tmap import _utils as tmap_util
    from cospar import help_functions as hf
    from cospar import logging as logg
    from cospar import settings
    from cospar import tool as tl

    current_folder = current_folder_input

    if 1:
        sample_name = sample_name
        print("Sample: ",sample_name)
        current_dir = os.getcwd()
        print("Changing the working directory from:", current_dir)
        #eng.run()
        loading_directory = os.path.join(current_folder, sample_name)
        kl_data_menu = "data"
        data_directory = os.path.join(loading_directory, kl_data_menu)
    
    print() 
    print("---------------------------------------------------------------------------------------")
    print("Please put the data at: ",loading_directory,"\data")
    print("---------------------------------------------------------------------------------------")
    print()
    print("Current working directory is:", data_directory)

    # 构建data.h5ad文件的完整路径
    h5ad_file_path = os.path.join(data_directory, data_name)
    adata_using = anndata.read_h5ad(h5ad_file_path)
    import scipy.sparse as sp
    # 将adata.X转换为稀疏矩阵类型
    sparse_X = sp.csr_matrix(adata_using.X.astype(float))
    adata_using.X=sparse_X
    import scipy.sparse as sp


    distance_matrix = matrix.distance_matrix(
        adata_using,
        round_of_smooth,
        neighbor_N,
        beta,
        truncation_threshold,
        save_subset,
        use_existing_KNN_graph=False,
        compute_new_Smatrix=True,
        use_full_Smatrix = True,
    )




    dense_matrix = distance_matrix.toarray()

    if sp.issparse(dense_matrix):
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
        print("!!!ERROR!!!: adata.X is changed as sparse matrix")
    else:
        print("adata.X is not a sparse matrix")


    if if_draw == "draw":
       
        import seaborn as sns
        import matplotlib.pyplot as plt
        import numpy as np
        from scipy import sparse
        # 假设 similarity_matrix 是你的稀疏矩阵数据
        # 将稀疏矩阵转换为非稀疏矩阵
        
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues')
        # 添加标题和标签
        plt.title('Distance Matrix')
        plt.xlabel('Samples')
        plt.ylabel('Samples')
        # 显示热图
        plt.show()
        import seaborn as sns
        # 假设你已经有了一个稠密矩阵 dense_matrix
        # 设置热图的上下限
        vmin = 0  # 最小值
        vmax = 0.004  # 最大值
        # 绘制热图
        sns.heatmap(dense_matrix, cmap='Blues', vmin=vmin, vmax=vmax)
        # 显示图形
        plt.show()


    return adata_using,loading_directory,dense_matrix





#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'LittleSnowFox'
DESCRIPTION = """KAILINss - TOGGLE(KAILIN): Single Cell Fate Tracing tools"""
URL = 'https://github.com/FullBlackWolf/KAILINss'
EMAIL = 'drchenjunpeng@hotmail.com'
AUTHOR = 'Junpeng Chen(Donald Louis)'
REQUIRES_PYTHON = ">=3.9, <3.10"
VERSION = '0.2.3.3'

# What packages are required for this module to be executed?
REQUIRED = [
    'anndata==0.10.6',
    'anyio==4.3.0',
    'argon2-cffi==23.1.0',
    'argon2-cffi-bindings==21.2.0',
    'array_api_compat==1.4.1',
    'arrow==1.3.0',
    'asttokens==2.4.1',
    'async-lru==2.0.4',
    'attrs==23.2.0',
    'Babel==2.14.0',
    'backports.tarfile==1.2.0',
    'beautifulsoup4==4.12.3',
    'bleach==6.1.0',
    'certifi==2024.2.2',
    'cffi==1.16.0',
    'charset-normalizer==3.3.2',
    'colorama==0.4.6',
    'comm==0.2.2',
    'contourpy==1.2.0',
    'cospar==0.3.3',
    'cycler==0.12.1',
    'debugpy==1.8.1',
    'decorator==5.1.1',
    'defusedxml==0.7.1',
    'docutils==0.21.2',
    'et_xmlfile==2.0.0',
    'ete3==3.1.3',
    'exceptiongroup==1.2.0',
    'executing==2.0.1',
    'fastcluster==1.2.6',
    'fastjsonschema==2.19.1',
    'fonttools==4.49.0',
    'fqdn==1.5.1',
    'frozendict==2.4.4',
    'get-annotations==0.1.2',
    'gseapy==1.1.2',
    'h11==0.14.0',
    'h5py==3.10.0',
    'html5lib==1.1',
    'httpcore==1.0.4',
    'httpx==0.27.0',
    'idna==3.6',
    'igraph==0.11.4',
    'importlib-metadata==6.8.0',
    'importlib_resources==6.3.0',
    'ipykernel==6.29.3',
    'ipython==8.18.1',
    'ipywidgets==8.1.2',
    'isoduration==20.11.0',
    'jaraco.classes==3.4.0',
    'jaraco.context==5.3.0',
    'jaraco.functools==4.0.1',
    'jedi==0.19.1',
    'Jinja2==3.1.3',
    'joblib==1.3.2',
    'json5==0.9.22',
    'jsonpointer==2.4',
    'jsonschema==4.21.1',
    'jsonschema-specifications==2023.12.1',
    'jupyter==1.0.0',
    'jupyter_client==8.6.1',
    'jupyter-console==6.6.3',
    'jupyter_core==5.7.2',
    'jupyter-events==0.9.1',
    'jupyter-lsp==2.2.4',
    'jupyter_server==2.13.0',
    'jupyter_server_terminals==0.5.3',
    'jupyterlab==4.1.4',
    'jupyterlab_pygments==0.3.0',
    'jupyterlab_server==2.25.4',
    'jupyterlab_widgets==3.0.10',
    'keyring==25.2.1',
    'kiwisolver==1.4.5',
    'lab==8.1',
    'leidenalg==0.10.2',
    'llvmlite==0.42.0',
    'lxml==5.2.2',
    'markdown-it-py==3.0.0',
    'MarkupSafe==2.1.5',
    'matplotlib==3.7.5',
    'matplotlib-inline==0.1.6',
    'mdurl==0.1.2',
    'mistune==3.0.1',
    'mizani==0.11.0',
    'more-itertools==10.2.0',
    'multitasking==0.0.11',
    'natsort==8.4.0',
    'nbclient==0.10.0',
    'nbconvert==7.8.0',
    'nbformat==5.10.2',
    'nest-asyncio==1.6.0',
    'networkx==3.2.1',
    'nh3==0.2.17',
    'notebook==7.1.1',
    'notebook_shim==0.2.4',
    'numba==0.59.0',
    'numpy==1.26.4',
    'openpyxl==3.1.5',
    'overrides==7.7.0',
    'packaging==24.0',
    'pandas==2.2.2',
    'pandocfilters==1.5.1',
    'parso==0.8.3',
    'patsy==0.5.6',
    'peewee==3.17.5',
    'pillow==10.2.0',
    'pip==24.2',
    'pkginfo==1.11.0',
    'platformdirs==4.2.0',
    'plotnine==0.13.1',
    'prometheus_client==0.20.0',
    'prompt-toolkit==3.0.43',
    'psutil==5.9.8',
    'pure-eval==0.2.2',
    'pycparser==2.21',
    'pygame==2.5.2',
    'Pygments==2.17.2',
    'pynndescent==0.5.11',
    'pyparsing==3.1.2',
    'python-dateutil==2.9.0.post0',
    'python-json-logger==2.0.7',
    'pytz==2024.1',
    #'pywin32==306',
    'pywin32-ctypes==0.2.2',
    'pywinpty==2.0.13',
    'PyYAML==6.0.1',
    'pyzmq==25.1.2',
    'qtconsole==5.5.1',
    'QtPy==2.4.1',
    'readme_renderer==43.0',
    'referencing==0.33.0',
    'requests==2.31.0',
    'requests-toolbelt==1.0.0',
    'rfc3339-validator==0.1.4',
    'rfc3986==2.0.0',
    'rfc3986-validator==0.1.1',
    'rich==13.7.1',
    'rpds-py==0.18.0',
    'scanpy==1.9.8',
    'scikit-learn==1.4.1.post1',
    'scikit-misc==0.3.1',
    'scipy==1.12.0',
    'seaborn==0.13.2',
    'Send2Trash==1.8.2',
    'session-info==1.0.0',
    'setuptools==68.2.2',
    'simplejson==3.19.2',
    'six==1.16.0',
    'sniffio==1.3.1',
    'soupsieve==2.5',
    'stack-data==0.6.3',
    'statsmodels==0.14.1',
    'stdlib-list==0.10.0',
    'terminado==0.18.1',
    'texttable==1.7.0',
    'threadpoolctl==3.3.0',
    'tinycss2==1.2.1',
    'tomli==2.0.1',
    'tornado==6.4',
    'tqdm==4.66.2',
    'traitlets==5.14.2',
    'twine==5.1.0',
    'txt2tags==3.9',
    'types-python-dateutil==2.8.19.20240311',
    'typing_extensions==4.10.0',
    'tzdata==2024.1',
    'umap-learn==0.5.5',
    'uri-template==1.3.0',
    'urllib3==2.2.1',
    'wcwidth==0.2.13',
    'webcolors==1.13',
    'webencodings==0.5.1',
    'websocket-client==1.7.0',
    'wheel==0.41.2',
    'widgetsnbextension==4.0.10',
    'yfinance==0.2.40',
    'zipp==3.18.0',
    'googledrivedownloader==0.4'
]


# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

import os
import io
from shutil import rmtree
import sys
from setuptools import Command

# Use relative path for current directory
try:
    with io.open('README.md', encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree('dist')  # Changed to relative path
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'], 

    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license='MIT',
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # package_data={
    #     'kailin': ['*', '*/*', '*/*/*', '*/*/*/*'],  # Include all files and subdirectories
    # },
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



import os
from pathlib import Path, PurePath

from scanpy import read

from . import logging as logg
from . import settings

url_prefix_0 = "https://kleintools.hms.harvard.edu/tools/downloads/cospar"


def synthetic_bifurcation(data_des="bifur"):
    """
    Synthetic clonal dataset with static barcoding.

    We simulated a differentiation process over a bifurcation fork.
    Cells are barcoded in the beginning, and the barcodes remain unchanged.
    In the simulation, we resample clones over time,
    like the experimental design to obtain the hematopoietic dataset
    or the reprogramming dataset. The dataset has two time points.

    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    # data_name='bifurcation_static_BC_adata_preprocessed.h5ad'
    data_name = "bifur_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


def raw_data_for_import_exercise():
    """
    The test dataset for demonstrating how to import your own data

    """

    url = "https://github.com/ShouWenWang-Lab/cospar/files/12036732/test_data.zip"
    path = "test_data.zip"
    path = Path(path)
    _check_datafile_present_and_download(path, backup_url=url)

    import zipfile

    with zipfile.ZipFile(f"test_data.zip", "r") as zip_ref:
        zip_ref.extractall()


# def synthetic_bifurcation_dynamic_BC(data_des='bifur_conBC'):
#     """
#     Synthetic clonal dataset with dynamic barcoding.

#     We simulated a differentiation process over a bifurcation fork.
#     Cells are barcoded, and the barcodes could accumulate mutations, which we call
#     `dynamic barcoding`. In the simulation, we resample clones over time,
#     like the experimental design to obtain the hematopoietic dataset
#     or the reprogramming dataset. The dataset has two time points.

#     Parameters
#     ----------
#     data_des: `str`
#         A key to label this dataset.
#     """

#     data_path=settings.data_path
#     figure_path=settings.figure_path
#     data_name='bifurcation_dynamic_BC_adata_preprocessed.h5ad'
#     return load_data_core(data_path,figure_path,data_name,data_des)


def reprogramming(data_des="CellTagging"):
    """
    The reprogramming dataset from

    * Biddy, B. A. et al. `Single-cell mapping of lineage and identity in direct reprogramming`. Nature 564, 219–224 (2018).

    This dataset has multiple time points for both the clones and the state measurements.
    The cells are barcoded over 3 rounds during the entire differentiation process.
    There are multiple ways to assemble the barcodes on day 0, day 3, and day 13
    into a clonal ID. Below, we provide three variants:

    * Concatenate barcodes on day 0 and day 13, as in the original
      analysis (adata.obsm['X_clone_Concat_D0D3'], the default);

    * Concatenate barcodes on day 0, day 3, and day 13 (adata.obsm['X_clone_Concat_D0D3D13']);

    * No concatenation; each cell has up to 3 barcodes (adata.obsm['X_clone_NonConcat_D0D3D13']).

    The last choice keeps the nested clonal structure in the data.
    You can choose any one of the clonal arrangement for downstream analysis,
    by setting `adata_orig.obsm['X_clone']=adata_orig.obsm['X_clone_Concat_D0D3']`.
    The three clonal arrangements give very similar fate prediction.

    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    data_name = "CellTagging_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


def reprogramming_Day0_3_28(data_des="Reprog_128"):
    """
    The reprogramming dataset from

    * Biddy, B. A. et al. `Single-cell mapping of lineage and identity in direct reprogramming`. Nature 564, 219–224 (2018).

    This dataset has time points on day 0, day 3, and day 28. Only day 28 has clonal information.
    The cells are barcoded over 3 rounds during the entire differentiation process.
    There are multiple ways to assemble the barcodes on day 0, day 3, and day 13
    into a clonal ID. Below, we concatenate barcodes on day 0 and day 13,
    as in the original analysis.


    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    data_name = "Reprog_128_D0D3_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


# def reprogramming_static_BC(data_des='CellTagging'):
#     """
#     The reprogramming dataset from

#     * Biddy, B. A. et al. `Single-cell mapping of lineage and identity in direct reprogramming`. Nature 564, 219–224 (2018).

#     This dataset has multiple time points for both the clones and the state measurements.

#     The cells are barcoded over 3 rounds during the entire differentiation process.
#     We combine up to 3 tags from the same cell into a single clonal label in representing
#     the X_clone matrix. In this representation, each cell has at most one clonal label.
#     Effectively, we convert the barcodes into static labels that do not carry temporal information.

#     Parameters
#     ----------
#     data_des: `str`
#         A key to label this dataset.
#     """

#     data_path=settings.data_path
#     figure_path=settings.figure_path
#     data_name='CellTagging_ConcatenateClone_adata_preprocessed.h5ad'
#     return load_data_core(data_path,figure_path,data_name,data_des)

# def reprogramming_dynamic_BC(data_des='CellTagging_NoConcat'):
#     """
#     The reprogramming dataset from

#     * Biddy, B. A. et al. `Single-cell mapping of lineage and identity in direct reprogramming`. Nature 564, 219–224 (2018).

#     This dataset has multiple time points for both the clones and the state measurements.

#     The cells are barcoded over 3 rounds during the entire differentiation process.
#     We treat barcode tags from each round as independent clonal label here. In this
#     representation, each cell can have multiple clonal labels at different time points.

#     Parameters
#     ----------
#     data_des: `str`
#         A key to label this dataset.
#     """

#     data_path=settings.data_path
#     figure_path=settings.figure_path
#     data_name='CellTagging_NoConcat_adata_preprocessed.h5ad'
#     return load_data_core(data_path,figure_path,data_name,data_des)


def lung(data_des="Lung"):
    """
    The direct lung differentiation dataset from

    * Hurley, K. et al. Cell Stem Cell (2020) doi:10.1016/j.stem.2019.12.009.

    This dataset has multiple time points for the state manifold, but only one time point
    for the clonal observation on day 27.

    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    data_name = "Lung_pos17_21_D27_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


def hematopoiesis(data_des="LARRY"):
    """
    The hematopoiesis data set from

    * Weinreb, C., Rodriguez-Fraticelli, A., Camargo, F. D. & Klein, A. M. Science 367, (2020)

    This dataset has 3 time points for both the clonal and state measurements. It only contains cells with clonal
    labels, which has ~50000 cells. Running the whole pipeline for the first time could take several hours in
    a standard personal computer.

    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    data_name = "LARRY_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


def hematopoiesis_130K(data_des="LARRY"):
    """
    The hematopoiesis data set from

    * Weinreb, C., Rodriguez-Fraticelli, A., Camargo, F. D. & Klein, A. M. Science 367, (2020)

    This dataset has 3 time points for both the clonal and state measurements. It includes
    cells with or without clonal labels, which has a total of ~130K cells.

    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    data_name = "Complete_LARRY_dataset_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


def hematopoiesis_Gata1_states(data_des="LARRY_Gata1_lineage"):
    """
    All of the hematopoiesis data set from

    * Weinreb, C., Rodriguez-Fraticelli, A., Camargo, F. D. & Klein, A. M. Science 367, (2020)

    This dataset includes non-clonally-labeled states that express Gata1. In total, it has ~38K cells.

    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    data_name = "LARRY_Gata1_lineage_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


def hematopoiesis_subsampled(data_des="LARRY_sp500_ranking1"):
    """
    Top 15% most heterogeneous clones of the hematopoiesis data set from

    * Weinreb, C., Rodriguez-Fraticelli, A., Camargo, F. D. & Klein, A. M. Science 367, (2020)

    This dataset has 3 time points for both the clones and the state measurements.
    This sub-sampled data better illustrates the power of CoSpar in robstly
    inferring differentiation dynamics from a noisy clonal dataset. Also, it
    is smaller and thus much faster to analyze.

    Parameters
    ----------
    data_des: `str`
        A key to label this dataset.
    """

    data_path = settings.data_path
    figure_path = settings.figure_path
    data_name = "LARRY_sp500_ranking1_adata_preprocessed.h5ad"
    return load_data_core(data_path, figure_path, data_name, data_des)


def load_data_core(
    data_path, figure_path, data_name, data_des, url_prefix=url_prefix_0
):
    url = f"{url_prefix}/{data_name}"
    path = os.path.join(data_path, data_name)
    path = Path(path)
    figure_path = Path(figure_path)

    if not path.parent.is_dir():
        logg.info(f"creating directory {path.parent} for saving data")
        path.parent.mkdir(parents=True)

    if not figure_path.is_dir():
        logg.info(f"creating directory {figure_path} for saving figures")
        figure_path.mkdir(parents=True)

    # print(url)
    status = _check_datafile_present_and_download(path, backup_url=url)
    if status:
        adata = read(path)
        # adata.uns['data_path']=[str(data_path)]
        # adata.uns['figure_path']=[str(figure_path)]
        adata.uns["data_des"] = [str(data_des)]
        return adata
    else:
        logg.error("Error, files do not exist")
        return None


def _check_datafile_present_and_download(path, backup_url=None):
    """Check whether the file is present, otherwise download."""
    path = Path(path)
    if path.is_file():
        return True
    if backup_url is None:
        return False
    logg.info(
        f"try downloading from url\n{backup_url}\n"
        "... this may take a while but only happens once"
    )
    if not path.parent.is_dir():
        logg.info(f"creating directory {path.parent} for saving data")
        path.parent.mkdir(parents=True)

    _download(backup_url, path)
    return True


def _download(url: str, path: Path):
    try:
        import ipywidgets
        from tqdm.auto import tqdm
    except ImportError:
        from tqdm import tqdm

    from urllib.request import Request, urlopen

    blocksize = 1024 * 8
    blocknum = 0

    try:
        with urlopen(Request(url, headers={"User-agent": "scanpy-user"})) as resp:
            total = resp.info().get("content-length", None)
            with tqdm(
                unit="B",
                unit_scale=True,
                miniters=1,
                unit_divisor=1024,
                total=total if total is None else int(total),
            ) as t, path.open("wb") as f:
                block = resp.read(blocksize)
                while block:
                    f.write(block)
                    blocknum += 1
                    t.update(len(block))
                    block = resp.read(blocksize)

    except (KeyboardInterrupt, Exception):
        # Make sure file doesn’t exist half-downloaded
        if path.is_file():
            path.unlink()
        raise




def choosemode_kl(parent_directory_origin,choosemode,debug):
    import os
    global empty_key
    #global current_folder

    current_folder = parent_directory_origin
 

    if choosemode ==  'Lineage':
        print('Current mode is Lineage tracing')
        folder_name = "database"
        new_directory = os.path.join(current_folder, folder_name)
        os.chdir(new_directory)
        current_folder = os.getcwd()
        print("Current folder:", current_folder)
        folder_name = "Tracing_sample"
        new_directory = os.path.join(current_folder, folder_name)
        os.chdir(new_directory)
        current_folder = os.getcwd()
        print("Current folder:", current_folder)
        



    if choosemode ==  'Clustering':
        print('Current mode is Lineage tracing')
        folder_name = "database"
        new_directory = os.path.join(current_folder, folder_name)
        os.chdir(new_directory)
        current_folder = os.getcwd()
        print("Current folder:", current_folder)
        folder_name = "Clustering_sample"
        new_directory = os.path.join(current_folder, folder_name)
        os.chdir(new_directory)
        current_folder = os.getcwd()
        print("Current folder:", current_folder)

    print('Current' + choosemode + ' sample list:')
    print(os.listdir(current_folder))
        
    return current_folder
    empty_key = 1


#需要增加：存稀疏矩阵还是非稀疏矩阵

def kl_save(
    Rep_data_folder,
    choosen_sample_name,
    distance_matrix,
    save_list,
    orig_adata
    ):

    import pandas as pd
    choosen_sample = choosen_sample_name
    print(Rep_data_folder)
    import scipy.io
    import anndata
    
    import os
    # 检查当前路径是否存在 'result' 文件夹
    if not os.path.exists('result'):
    # 如果不存在，则创建 'result' 文件夹
        os.makedirs('result')
    
    folder_name = "result"
    result_directory = os.path.join(Rep_data_folder, folder_name)
    
   
    '''
    for item in save_list:
        save_csv_sample = pd.DataFrame(item)
        save_csv_filename = item + ".csv"
        save_csv_sample.to_csv(save_csv_filename, index=False)

        
         #这里要加一个标题检测，即第一行不是label时，需要自动位移。此处设计成操作限定的形式，即matlab只认第一行的固定标签字样
         #并且只认固定的格式，不然就报错，避免导致错误输入。
    '''


    merged_csv = pd.DataFrame()

    for item in save_list:
        # 使用 eval 动态评估字符串以获取相应的数据
        data = eval(item)
        print(item)
        df = pd.DataFrame(data)
        if 'obsm' in item:
            df = pd.DataFrame(data)
        elif 'obs' in item:
            df = pd.DataFrame(data).reset_index()

        merged_csv = pd.concat([merged_csv, df], axis=1)


        
    
    # 生成变量名列表
    var_names = [f"Var{i+1}" for i in range(len(merged_csv.columns))]

    # 重命名列
    merged_csv.columns = var_names


    folder_name = "result"
    result_directory = os.path.join(Rep_data_folder, folder_name)
    mat_file_path = os.path.join(result_directory, 'distance_matrix.mat')
    scipy.io.savemat(mat_file_path, {'distance_matrix': distance_matrix})
    merged_csv.to_csv(os.path.join(result_directory, "merged_data.csv"), index=False)

    # 输出 DataFrame
    print(merged_csv.head())  # 打印前几行查看数据



    return merged_csv,result_directory




def kl_autosave(
    Rep_data_folder,
    choosen_sample_name,
    distance_matrix,
    save_list,
    orig_adata,
    round_of_smooth,
    neighbor_N
    ):

    import pandas as pd
    choosen_sample = choosen_sample_name
    print(Rep_data_folder)
    import scipy.io
    import anndata
    
    import os
    # 检查当前路径是否存在 'result' 文件夹
    if not os.path.exists('result'):
    # 如果不存在，则创建 'result' 文件夹
        os.makedirs('result')
    
    folder_name = "result"
    result_directory = os.path.join(Rep_data_folder, folder_name)
    
   
    '''
    for item in save_list:
        save_csv_sample = pd.DataFrame(item)
        save_csv_filename = item + ".csv"
        save_csv_sample.to_csv(save_csv_filename, index=False)

        
         #这里要加一个标题检测，即第一行不是label时，需要自动位移。此处设计成操作限定的形式，即matlab只认第一行的固定标签字样
         #并且只认固定的格式，不然就报错，避免导致错误输入。
    '''


    merged_csv = pd.DataFrame()

    for item in save_list:
        # 使用 eval 动态评估字符串以获取相应的数据
        data = eval(item)
        print(item)
        df = pd.DataFrame(data)
        if 'obsm' in item:
            df = pd.DataFrame(data)
        elif 'obs' in item:
            df = pd.DataFrame(data).reset_index()

        merged_csv = pd.concat([merged_csv, df], axis=1)


        
    
    # 生成变量名列表
    var_names = [f"Var{i+1}" for i in range(len(merged_csv.columns))]

    # 重命名列
    merged_csv.columns = var_names

    folder_name = "result"
    result_directory = os.path.join(Rep_data_folder, folder_name)
    mat_file_path = os.path.join(result_directory, 'r' + str(round_of_smooth) + 'n' + str(neighbor_N) + 'distance_matrix.mat')
    scipy.io.savemat(mat_file_path, {'distance_matrix': distance_matrix})
    merged_csv.to_csv(os.path.join(result_directory, 'r' + str(round_of_smooth) + 'n' + str(neighbor_N) + "merged_data.csv"), index=False)

    # 输出 DataFrame
    print(merged_csv.head())  # 打印前几行查看数据



    return merged_csv,result_directory
